# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2018 The Google AI Language Team Authors and The HugginFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import math

import torch
from torch import nn

from nemo import logging
from nemo.collections.nlp.utils.functional_utils import gelu

__all__ = []


try:
    from apex.normalization import FusedLayerNorm

    # Try to use FusedLayerNorm from Apex - this will trigger an error.
    _ = FusedLayerNorm(8, eps=1e-5)

except Exception as e:
    logging.warning("Unable to import FusedLayerNorm  from APEX. Using regular LayerNorm instead.")
    from torch.nn import LayerNorm as FusedLayerNorm


class FixedPositionalEncoding(nn.Module):
    """
    Fixed positional encoding (embedding layer) from sine and cosine functions
    of different frequencies according to https://arxiv.org/abs/1706.03762

    Args:
        hidden_size: size of the embeddings in the model, also known as d_model
        max_sequence_length: maximum allowed length of the input sequence
    """

    def __init__(self, hidden_size, max_sequence_length=512):
        super().__init__()

        pos_enc = torch.zeros(max_sequence_length, hidden_size)
        position = torch.arange(0.0, max_sequence_length).unsqueeze(1)
        coef = -math.log(10000.0) / hidden_size
        div_term = torch.exp(coef * torch.arange(0.0, hidden_size, 2))
        pos_enc[:, 0::2] = torch.sin(position * div_term)
        pos_enc[:, 1::2] = torch.cos(position * div_term)
        pos_enc.div_(math.sqrt(hidden_size))
        self.register_buffer('pos_enc', pos_enc)

    def forward(self, position_ids):
        return torch.embedding(self.pos_enc, position_ids)


class TransformerEmbedding(nn.Module):
    """
    Embedding from token and position embeddings.
    Optionally add token_type embedding (e.g. type of the sentence in BERT).

    Args:
        vocab_size: size of the vocabulary
        hidden_size: size of the embeddings in the model, also known as d_model
        max_sequence_length: maximum allowed length of the input sequence
        num_token_types: number of different token types
            (e.g. tokens of sentence A and tokens of sentence B in BERT)
        embedding_dropout: probability of dropout applied to embeddings
    """

    def __init__(
        self,
        vocab_size,
        hidden_size,
        max_sequence_length=512,
        num_token_types=2,
        embedding_dropout=0.0,
        learn_positional_encodings=False,
    ):
        super().__init__()

        self.max_sequence_length = max_sequence_length
        self.token_embedding = nn.Embedding(vocab_size, hidden_size, padding_idx=0)
        if learn_positional_encodings:
            self.position_embedding = nn.Embedding(max_sequence_length, hidden_size)
        else:
            self.position_embedding = FixedPositionalEncoding(hidden_size, max_sequence_length)
        self.token_type_embedding = nn.Embedding(num_token_types, hidden_size)
        self.layer_norm = FusedLayerNorm(hidden_size, eps=1e-5)
        self.dropout = nn.Dropout(embedding_dropout)

    def forward(self, input_ids, token_type_ids=None, start_pos=0):
        seq_length = input_ids.size(1)
        if seq_length > self.max_sequence_length:
            raise ValueError("Input sequence is longer than maximum allowed sequence length for positional encoding")
        position_ids = torch.arange(
            start=start_pos, end=start_pos + seq_length, dtype=torch.long, device=input_ids.device
        )
        position_ids = position_ids.unsqueeze(0).expand_as(input_ids)

        token_embeddings = self.token_embedding(input_ids)
        position_embeddings = self.position_embedding(position_ids)
        embeddings = token_embeddings + position_embeddings

        if token_type_ids is not None:
            token_type_embeddings = self.token_type_embedding(token_type_ids)
            embeddings = embeddings + token_type_embeddings

        embeddings = self.layer_norm(embeddings)
        embeddings = self.dropout(embeddings)

        return embeddings


class MultiHeadAttention(nn.Module):
    """
    Multi-head scaled dot-product attention layer.

    Args:
        hidden_size: size of the embeddings in the model, also known as d_model
        num_attention_heads: number of heads in multi-head attention
        attn_score_dropout: probability of dropout applied to attention scores
        attn_layer_dropout: probability of dropout applied to the output of the
            whole layer, but before layer normalization
    """

    def __init__(self, hidden_size, num_attention_heads, attn_score_dropout=0.0, attn_layer_dropout=0.0):
        super().__init__()
        if hidden_size % num_attention_heads != 0:
            raise ValueError(
                "The hidden size (%d) is not a multiple of the number "
                "of attention heads (%d)" % (hidden_size, num_attention_heads)
            )
        self.hidden_size = hidden_size
        self.num_attention_heads = num_attention_heads
        self.attn_head_size = int(hidden_size / num_attention_heads)
        self.attn_scale = math.sqrt(math.sqrt(self.attn_head_size))

        self.query_net = nn.Linear(hidden_size, hidden_size)
        self.key_net = nn.Linear(hidden_size, hidden_size)
        self.value_net = nn.Linear(hidden_size, hidden_size)
        self.out_projection = nn.Linear(hidden_size, hidden_size)

        self.attn_dropout = nn.Dropout(attn_score_dropout)
        self.layer_dropout = nn.Dropout(attn_layer_dropout)
        self.layer_norm = FusedLayerNorm(hidden_size, eps=1e-5)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attn_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(self, queries, keys, values, attention_mask):

        # attention_mask is needed to hide the tokens which correspond to [PAD]
        # in the case of BERT, or to hide the future tokens in the case of
        # vanilla language modeling and translation
        query = self.query_net(queries)
        key = self.key_net(keys)
        value = self.value_net(values)
        query = self.transpose_for_scores(query) / self.attn_scale
        key = self.transpose_for_scores(key) / self.attn_scale
        value = self.transpose_for_scores(value)

        # for numerical stability we pre-divide query and key by sqrt(sqrt(d))
        # and perform attention probs computation in float32
        attention_scores = torch.matmul(query, key.transpose(-1, -2)).float()
        if attention_mask is not None:
            attention_scores = attention_scores + attention_mask.float()
        attention_probs = torch.softmax(attention_scores, dim=-1).to(key.dtype)
        attention_probs = self.attn_dropout(attention_probs)

        context = torch.matmul(attention_probs, value)
        context = context.permute(0, 2, 1, 3).contiguous()
        new_context_shape = context.size()[:-2] + (self.hidden_size,)
        context = context.view(*new_context_shape)

        # output projection
        output_states = self.out_projection(context)
        output_states = self.layer_dropout(output_states)
        output_states = self.layer_norm(queries + output_states)

        return output_states


class LightweightConv1d(nn.Module):
    """
    Lightweight convolution layer from https://arxiv.org/abs/1901.10430

    Args:
        hidden_size: size of the embeddings in the model, also known as d_model
        num_heads: number of heads in lightweight convolution
        kernel_size: convolution kernel size
        conv_weight_dropout: probability of dropout applied to the convolution
            kernel (strictly speaking, DropConnect)
        conv_layer_dropout: probability of dropout applied to the output of the
            whole layer, but before layer normalization
    """

    def __init__(self, hidden_size, num_attention_heads, kernel_size, conv_weight_dropout=0.0, conv_layer_dropout=0.0):
        super().__init__()
        self.num_heads = num_attention_heads
        self.kernel_size = kernel_size
        self.weight = nn.Parameter(torch.Tensor(num_attention_heads, 1, kernel_size))
        self.in_projection = nn.Linear(hidden_size, hidden_size)
        self.out_projection = nn.Linear(hidden_size, hidden_size)

        self.conv_weight_dropout = nn.Dropout(conv_weight_dropout)
        self.conv_layer_dropout = nn.Dropout(conv_layer_dropout)
        self.layer_norm = FusedLayerNorm(hidden_size, eps=1e-5)

    def forward(self, hidden_states, attention_mask):
        batch_size, seq_len, hidden_size = hidden_states.size()
        output_states = self.in_projection(hidden_states)
        output_states = output_states.permute(0, 2, 1)

        weight = torch.softmax(self.weight, dim=-1)
        weight = self.conv_weight_dropout(weight)

        if attention_mask:
            pivot = self.kernel_size // 2 + 1
            weight[:, :, pivot:] = 0

        output_states = output_states.contiguous().view(-1, self.num_heads, seq_len)
        output_states = torch.conv1d(output_states, weight, padding=self.kernel_size // 2, groups=self.num_heads)
        output_states = output_states.view(batch_size, hidden_size, seq_len)
        output_states = output_states.permute(0, 2, 1)

        # output projection
        output_states = self.out_projection(output_states)
        output_states = self.conv_layer_dropout(output_states)
        output_states = self.layer_norm(hidden_states + output_states)

        return output_states


class TwoStreamSelfAttention(nn.Module):
    """
    Two-Stream Self-Attention layer from https://arxiv.org/abs/1906.08237

    Args:
        hidden_size: size of the embeddings in the model, also known as d_model
        num_attention_heads: number of heads in multi-head attention
        attn_score_dropout: probability of dropout applied to attention scores
        attn_layer_dropout: probability of dropout applied to the output of the
            whole layer, but before layer normalization
    """

    def __init__(self, hidden_size, num_attention_heads, attn_score_dropout=0.0, attn_layer_dropout=0.0):
        super().__init__()
        self.query_stream = MultiHeadAttention(
            hidden_size, num_attention_heads, attn_score_dropout, attn_layer_dropout
        )
        self.content_stream = MultiHeadAttention(
            hidden_size, num_attention_heads, attn_score_dropout, attn_layer_dropout
        )

    def forward(self, query_states, content_states, query_attention_mask, content_attention_mask):
        output_query_states = self.query_stream(query_states, content_states, content_states, query_attention_mask)
        output_content_states = self.content_stream(
            query_states, content_states, content_states, content_attention_mask
        )
        return output_query_states, output_content_states


class PositionWiseFF(nn.Module):
    """
    Position-wise feed-forward network of Transformer block.

    Args:
        hidden_size: size of the embeddings in the model, also known as d_model
        inner_size: number of neurons in the intermediate part of feed-forward
            net, usually is (4-8 x hidden_size) in the papers
        ffn_dropout: probability of dropout applied to net output
        hidden_act: activation function used between two linear layers
    """

    def __init__(self, hidden_size, inner_size, ffn_dropout=0.0, hidden_act="relu"):
        super().__init__()
        self.dense_in = nn.Linear(hidden_size, inner_size)
        self.dense_out = nn.Linear(inner_size, hidden_size)
        self.layer_dropout = nn.Dropout(ffn_dropout)
        self.layer_norm = FusedLayerNorm(hidden_size, eps=1e-5)
        ACT2FN = {"gelu": gelu, "relu": torch.relu}
        self.act_fn = ACT2FN[hidden_act]

    def forward(self, hidden_states):
        output_states = self.dense_in(hidden_states)
        output_states = self.act_fn(output_states)
        output_states = self.dense_out(output_states)
        output_states = self.layer_dropout(output_states)
        output_states = self.layer_norm(hidden_states + output_states)
        return output_states



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2019, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import List, Optional, Union

from nemo.core.neural_factory import Optimization
from nemo.core.neural_modules import ModuleType
from nemo.core.neural_types import NmTensor
from nemo.utils.app_state import AppState


def topological_sort_from_leaves(leaf_nmtensors: List[NmTensor], cached_training_state: 'TrainingState' = None):
    """A function that accepts a list of NmTensors that need to be computed and constructs a callchain DAG that starts
    from a datalayerNM and can be used to compute the NmTensors.

    args:
        leaf_nmtensors (List[NmTensors]): The tensors to be computed
        cached_training_state (TrainingState): A dictionary of already computed tensors.
            Defaults to None meaning an empty cache.

    returns:
        top_sorted_modules: the callchain DAG
    """

    def create_node(producer, producer_args):
        if producer_args is None:
            return tuple((producer, ()))
        return tuple((producer, tuple([(k, v) for k, v in producer_args.items()]),))

    def is_in_degree_zero(node, processed_nodes, cached_training_state):
        """A node has in degree of zero"""
        if node[1] == ():
            return True
        for _, nmtensor in node[1]:
            node = create_node(nmtensor.producer, nmtensor.producer_args)
            if node not in processed_nodes:
                if cached_training_state and cached_training_state.check_tensor_cached(nmtensor.unique_name):
                    continue
                return False
        return True

    hooks = leaf_nmtensors if isinstance(leaf_nmtensors, list) else [leaf_nmtensors]

    # ensures that no tensors are processed twice
    processed_nmtensors = set()

    indices_to_remove = []
    # Check for duplicates in hook
    for i, nmtensor in enumerate(hooks):
        if nmtensor in processed_nmtensors:
            indices_to_remove.append(i)
        else:
            processed_nmtensors.add(nmtensor)

    for i in reversed(indices_to_remove):
        hooks.pop(i)

    _top_sorted_modules = []
    all_nodes = {}

    # extract all nodes to all_nodes set
    hooks_lst = list(hooks)
    while len(hooks_lst) > 0:
        # take nmtensor from the end of the list
        nmtensor = hooks_lst.pop()
        producer_args = nmtensor.producer_args

        node = create_node(nmtensor.producer, producer_args)
        # Store nmtensor as an output of its producer
        # first make sure all keys are present per output port
        # and nm is inside all_nodes
        if node not in all_nodes:
            all_nodes[node] = {k: None for k in nmtensor.producer.output_ports}
        # second, populate output port with current nmtensor
        # where applicable
        all_nodes[node][nmtensor.output_port_name] = nmtensor
        processed_nmtensors.add(nmtensor)

        new_tensors = set()
        if producer_args is not None and producer_args != {}:
            for _, new_nmtensor in producer_args.items():
                if new_nmtensor not in processed_nmtensors:
                    new_tensors.add(new_nmtensor)

        if cached_training_state:
            for _, input_nmtensor in producer_args.items():
                if cached_training_state.check_tensor_cached(input_nmtensor.unique_name):
                    new_tensors.remove(input_nmtensor)

        # Order the new set so that all ranks have the callchain in the same order
        new_tensors = sorted(list(new_tensors), key=lambda x: str(x))
        for new_nmtensor in new_tensors:
            # put in the start of list
            hooks_lst.insert(0, new_nmtensor)

    all_node_with_output = []
    # Iterate over all_nodes to create new nodes that include its output
    # now all nodes have (module, input tensors, output tensors)
    for node in all_nodes:
        all_node_with_output.append(tuple((node[0], node[1], all_nodes[node])))

    processed_nodes = []
    while len(all_node_with_output) > 0:
        for node in all_node_with_output.copy():
            # if node's in_degree is zero it can be added to
            # _top_sorted_modules
            # this will also reduce in_degree of its children
            if is_in_degree_zero(node, processed_nodes, cached_training_state):
                _top_sorted_modules.append(node)
                processed_nodes.append((node[0], node[1]))
                all_node_with_output.remove(node)

    # Create top_sorted_modules aka callchain
    top_sorted_modules = []
    for i, mod in enumerate(_top_sorted_modules):
        top_sorted_modules.append((mod[0], dict(mod[1]), mod[2]))
        # Ensure that there is only one dataset in callchain
        if i > 0 and mod[0].type == ModuleType.datalayer:
            raise ValueError("There were more than one DataLayer NeuralModule inside your DAG.")

        if cached_training_state and mod[0].type == ModuleType.datalayer:
            raise ValueError("Could not compute tensor from current cached training state.")

    return top_sorted_modules


class TrainingState:
    def __init__(self, action: 'Actions'):
        """A class used to wrap the current training state of an Actions.train() function. This class holds a mapping
        of tensor.unique_name -> it's backend tensor (eg Pytorch Tensor) or None if the tensor has been been computed
        on the current step.

        args:
            action (Actions): The Actions object this state is associated with.
        """
        tensor_naming_registery = AppState().tensor_names
        self.tensor_dict = {}.fromkeys(tensor_naming_registery.unique_names, None)
        self._action = action

    def tensor_list(self):
        """Returns a list the unique names of all tensors.
        """
        return self.tensor_dict.keys()

    def clear_dict(self):
        """Clears the dictionary by setting all values to None. Used in-between training batches to clear it's state.
        """
        for name in self.tensor_dict:
            self.tensor_dict[name] = None

    def set_tensor(self, tensor: NmTensor, value: 'torch.Tensor'):
        """Sets the value of tensor

        args:
            tensor (NmTensor)
            value (torch.Tensor)
        """
        self.tensor_dict[tensor.unique_name] = value

    def check_tensor_cached(self, unique_name: str):
        """Checks to see the tensor value has been computed in the current step yet.

        args:
            unique_name (str): The NmTensor.unique_name that we want to check for.

        returns:
            (bool) whether the tensor with unique_name has been computed yet.
        """
        if self.tensor_dict[unique_name] is None:
            return False
        return True

    def get_tensor(self, name: Union[str, NmTensor], compute: bool = True):
        """Returns the value associated with a tensor. And optionally, computes the value of the tensor if not already
        set.

        args:
            name (str, NmTensor): The user-defined name for a tensor or the NmTensor itself.
            compute (bool): If True and the tensor has not already been computed, there will be an attempt to create a
                call DAG and then do a forward pass on this call DAG to compute the tensor. If False, it will return
                None if the tensor has not been computed yet.
                Defaults to True.

        returns:
            (torch.tensor or None) representing the computed value of the requested name. Returns None if compute is
            False and the tensor has not been computed yet.
        """
        if isinstance(name, NmTensor):
            unique_name = name.unique_name
        else:
            unique_name = AppState().tensor_names[name]
        tensor_value = self.tensor_dict[unique_name]
        if tensor_value is None and compute:
            nmtensor = AppState().tensor_names._nmtensor_uniname_dict[unique_name]
            callchain = topological_sort_from_leaves([nmtensor], cached_training_state=self)
            callchain.insert(0, ())
            self._action.nm_graph_forward_pass(callchain, self.tensor_dict)
            tensor_value = self.tensor_dict[unique_name]
        return tensor_value


class Actions(ABC):
    """Basic actions allowed on graphs of Neural Modules"""

    def __init__(self, local_rank, global_rank, optimization_level=Optimization.mxprO0):
        self._local_rank = local_rank
        self._global_rank = global_rank
        self._optim_level = optimization_level

    @property
    def local_rank(self):
        """Local rank during distributed execution. None if single GPU/CPU

        Returns:
            (int) rank or worker or None if not in distributed model
        """
        return self._local_rank

    @property
    def global_rank(self):
        """Global rank during distributed execution. None if single GPU/CPU

        Returns:
            (int) rank or worker or None if not in distributed model
        """
        return self._global_rank

    @abstractmethod
    def train(
        self,
        tensors_to_optimize: List[NmTensor],
        callbacks: Optional[List[Union['ActionCallback', 'NeMoCallback']]],
        lr_policy=None,
        batches_per_step=None,
        stop_on_nan_loss=False,
    ):
        """This action executes training and (optionally) evaluation.

        Args:
            tensors_to_optimize: which tensors to optimize. Typically this is
                single loss tesnor.
            callbacks: list of callback objects
            lr_policy: function which should take (initial_lr, step, epoch) and
                return learning rate
            batches_per_step: number of mini-batches to process before one
                optimizer step. (default: None, same as 1). Use this
                to simulate larger batch sizes on hardware which could not fit
                larger batch in memory otherwise. Effectively, this will make
                "algorithmic" batch size per GPU/worker = batches_per_step*
                batch_size
            stop_on_nan_loss: (default: False) If set to True, the training
                will stop if loss=nan or inf. If set to False, the training
                will continue.

        Returns:
            None
        """
        pass

    @abstractmethod
    def infer(self, tensors: List[NmTensor]):
        """This action executes inference. Nothing is optimized.
        Args:
          tensors: which tensors to evaluate.

        Returns:
          None
        """
        pass

    @abstractmethod
    def create_optimizer(self, optimizer, things_to_optimize, optimizer_params):
        """
        Creates an optimizer object to be use in the train() method.

        Args:
            optimizer: Specifies which optimizer to use.
            things_to_optimize: A list of neural modules or tensors to be
                optimized.
            optimizer_params: Specifies the parameters of the optimizer

        Returns:
            Optimizer
        """
        pass



import torch
from torch.utils.data import Dataset

from ....core.neural_types import *
from ...pytorch.nm import DataLayerNM


def neuralType2TensorShape(neural_type: NeuralType, default_dim=32, skip_batch_axis=True) -> torch.Size:
    """
    Converts Neural Type to torch tensor shape.
    Args:
      neural_type: input Neural Type
      default_dim: default dimension to use if not specified by Neural Type
      skip_batch_axis: (default: True) whether to skip batch axis is resulting
      shape.

    Returns:
      torch.Size
    """
    dims = []
    for axis in neural_type.axes:
        if axis.kind == AxisKind.Batch and skip_batch_axis:
            continue
        if axis.size is not None:
            dims.append(axis.size)
        else:
            dims.append(default_dim)
    return torch.Size(dims)


class _ZeroDS(Dataset):
    def __init__(self, size, shapes, dtype):
        Dataset.__init__(self)
        self._size = size
        self._tensor_shapes = shapes
        self._type = dtype

    def __getitem__(self, index):
        if self._type is not None:
            if not isinstance(self._type, list):
                types = [self._type] * len(self._tensor_shapes)
            else:
                types = self._type
        else:
            types = [torch.FloatTensor] * len(self._tensor_shapes)
        res = []
        for ts, tp in zip(self._tensor_shapes, types):
            res.append(torch.zeros(ts).type(tp))
        return tuple(res)

    def __len__(self):
        return self._size


class ZerosDataLayer(DataLayerNM):
    """
    DataLayer Neural Module which emits zeros.
    This module should be used for debugging/benchmarking purposes.

    Args:
        size: (int) size of the underlying dataset
        output_ports: which output ports it should have
        dtype: Dtype of the output tensors.
        batch_size (int): Size of batches to output.
        shapes: If None, this will be inferred from output_ports. Else,
            specifies the shape of the output tensors.
            Defaults to None.
    """

    def __init__(self, size, output_ports, dtype, batch_size, shapes=None):
        self._output_ports = output_ports
        DataLayerNM.__init__(self)
        self._size = size
        self._type = dtype
        self._batch_size = batch_size
        self._shapes = shapes
        if self._shapes is None:
            self._shapes = [neuralType2TensorShape(pval) for pname, pval in self._output_ports.items()]

        self._dataset = _ZeroDS(size=self._size, shapes=self._shapes, dtype=self._type)

    @property
    def input_ports(self):
        return {}

    @property
    def output_ports(self):
        return self._output_ports

    def __len__(self):
        return len(self._dataset)

    @property
    def data_iterator(self):
        return None

    @property
    def dataset(self):
        return self._dataset



# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from weakref import WeakValueDictionary


class NmTensorNameRegistry:
    def __init__(self):
        """
            Constructor. Initializes the NmTensorNameRegistry. Reserves the default 'loss' name.

            TODO: We should be recording the tensors of each graph rather than all the tensors.
        """
        # Create the nmtensor_naming_dict
        # which contains a mapping of str to NMTensor.unique_name
        self._nmtensor_naming_dict = {"loss": "loss"}  # Reserve keyname of 'loss'
        # Create a dict that maps unique_names to tensors for use with TrainingState.get_tensor()
        self._nmtensor_uniname_dict = WeakValueDictionary()

    @property
    def unique_names(self):
        """Returns the set of all NmTensors.unique_names + 'loss'
        """
        return list(self._nmtensor_uniname_dict.keys()) + ["loss"]

    def register(self, tensor: 'NmTensor'):
        """Helper function to register a newly created NmTensor by adding it to self.__nmtensor_uniname_dict.
        Should be called from NmTensor.__init__()

        args:
            tensor (NmTensor): The tensor to be registered.
        """

        # Check if object is already in a set.
        if tensor.unique_name in self._nmtensor_uniname_dict:
            pass

        # Finally, add object to the set.
        self._nmtensor_uniname_dict[tensor.unique_name] = tensor

    def rename_NmTensor(self, tensor: 'NmTensor', new_name: str):
        """Helper function that changes the naming dictionary to facilitate user name -> tensor.unique_name lookup.

        args:
            tensor (NmTensor): The tensor to be renamed.
            new_name (str): its new name.
        """
        # Find old name if exists
        old_name = tensor.unique_name
        for custom_name, unique_name in self._nmtensor_naming_dict.items():
            if unique_name == tensor.unique_name:
                old_name = custom_name

        if old_name != tensor.unique_name:
            del self._nmtensor_naming_dict[old_name]

        if new_name in self._nmtensor_naming_dict:
            raise KeyError(f"{new_name} already exists in current graph. Please use a unique name")
        self._nmtensor_naming_dict[new_name] = tensor.unique_name

    def __getitem__(self, key: str):
        """
        Object getter function.

        Args:
            key: Object name.

        Returns:
            Object associated with the key.
        """
        # Search for an object with a given name.
        if key in self._nmtensor_naming_dict:
            key = self._nmtensor_naming_dict[key]

        if key in self._nmtensor_uniname_dict or key == "loss":
            return key

        raise KeyError("A NmTensor with name `{}` don't exists!".format(key))



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import numpy as np

from nemo import logging


def analyze_confusion_matrix(cm, dict, max_pairs=10):
    """
    Sort all confusions in the confusion matrix by value and display results.
    Print results in a format: (name -> name, value)
    Args:
        cm: Confusion matrix
        dict: Dictionary with key as a name and index as a value (Intents or Slots)
        max_pairs: Max number of confusions to print
    """
    threshold = 5  # just arbitrary value to take confusion with at least this number
    confused_pairs = {}
    size = cm.shape[0]
    for i in range(size):
        res = cm[i].argsort()
        for j in range(size):
            pos = res[size - j - 1]
            # no confusion - same row and column
            if pos == i:
                continue
            elif cm[i][pos] >= threshold:
                str = f'{dict[i]} -> {dict[pos]}'
                confused_pairs[str] = cm[i][pos]
            else:
                break

    # sort by max confusions and print first max_pairs
    sorted_confused_pairs = sorted(confused_pairs.items(), key=lambda x: x[1], reverse=True)
    for i, pair_str in enumerate(sorted_confused_pairs):
        if i >= max_pairs:
            break
        logging.info(pair_str)


def errors_per_class(cm, dict):
    """
    Summarize confusions per each class in the confusion matrix.
    It can be useful both for Intents and Slots.
    It counts each confusion twice in both directions.
    Args:
        cm: Confusion matrix
        dict: Dictionary with key as a name and index as a value (Intents or Slots)
    """
    size = cm.shape[0]
    confused_per_class = {}
    total_errors = 0
    for class_num in range(size):
        sum = 0
        for i in range(size):
            if i != class_num:
                sum += cm[class_num][i]
                sum += cm[i][class_num]
        confused_per_class[dict[class_num]] = sum
        total_errors += sum
        # logging.info(f'{dict[class_num]} - {sum}')

    logging.info(f'Total errors (multiplied by 2): {total_errors}')
    sorted_confused_per_class = sorted(confused_per_class.items(), key=lambda x: x[1], reverse=True)
    for conf_str in sorted_confused_per_class:
        logging.info(conf_str)


def log_misclassified_queries(intent_labels, intent_preds, queries, intent_dict, limit=50):
    """
    Display examples of Intent mistakes.
    In a format: Query, predicted and labeled intent names.
    """
    logging.info(f'*** Misclassified intent queries (limit {limit}) ***')
    cnt = 0
    for i in range(len(intent_preds)):
        if intent_labels[i] != intent_preds[i]:
            query = queries[i].split('\t')[0]
            logging.info(
                f'{query} (predicted: {intent_dict[intent_preds[i]]} - labeled: {intent_dict[intent_labels[i]]})'
            )
            cnt = cnt + 1
            if cnt >= limit:
                break


def log_misclassified_slots(
    intent_labels, intent_preds, slot_labels, slot_preds, subtokens_mask, queries, intent_dict, slot_dict, limit=50
):
    """
    Display examples of Slot mistakes.
    In a format: Query, predicted and labeled intent names and list of predicted and labeled slot numbers.
    also prints dictionary of the slots at the start for easier reading.
    """
    logging.info('')
    logging.info(f'*** Misclassified slots queries (limit {limit}) ***')
    # print slot dictionary
    logging.info(f'Slot dictionary:')
    str = ''
    for i, slot in enumerate(slot_dict):
        str += f'{i} - {slot}, '
        if i % 5 == 4 or i == len(slot_dict) - 1:
            logging.info(str)
            str = ''

    logging.info('----------------')
    cnt = 0
    for i in range(len(intent_preds)):
        cur_slot_pred = slot_preds[i][subtokens_mask[i]]
        cur_slot_label = slot_labels[i][subtokens_mask[i]]
        if not np.all(cur_slot_pred == cur_slot_label):
            query = queries[i].split('\t')[0]
            logging.info(
                f'{query} (predicted: {intent_dict[intent_preds[i]]} - labeled: {intent_dict[intent_labels[i]]})'
            )
            logging.info(f'p: {cur_slot_pred}')
            logging.info(f'l: {cur_slot_label}')
            cnt = cnt + 1
            if cnt >= limit:
                break


def check_problematic_slots(slot_preds_list, slot_dict):
    """ Check non compliance of B- and I- slots for datasets that use such slot encoding. """
    cnt = 0

    # for sentence in slot_preds:
    # slots = sentence.split(" ")
    sentence = slot_preds_list
    for i in range(len(sentence)):
        slot_name = slot_dict[int(sentence[i])]
        if slot_name.startswith("I-"):
            prev_slot_name = slot_dict[int(sentence[i - 1])]
            if slot_name[2:] != prev_slot_name[2:]:
                print("Problem: " + slot_name + " - " + prev_slot_name)
                cnt += 1
    print("Total problematic slots: " + str(cnt))



# Copyright (C) NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    'deprecated',
]

import functools

import wrapt

from nemo.utils import logging

# Remember which deprecation warnings have been printed already.
_PRINTED_WARNING = {}


def deprecated(wrapped=None, version=None, explanation=None):
    """ Decorator class used for indicating that a function is deprecated and going to be removed.
    Tracks down which functions printed the warning and will print it only once per function.
    """

    if wrapped is None:
        return functools.partial(deprecated, version=version, explanation=explanation)

    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        """
        Method prints the adequate warning (only once per function) when
        required and calls the function func, passing the original arguments,
        i.e. version and explanation.

        Args:
          version: Version in which the function will be removed (optional)
          explanation: Additional explanation (optional), e.g. use method ``blabla instead``.
        """

        # Check if we already warned about that function.
        if wrapped.__name__ not in _PRINTED_WARNING.keys():
            # Add to list so we won't print it again.
            _PRINTED_WARNING[wrapped.__name__] = True

            # Prepare the warning message.
            msg = "Function ``{}`` is deprecated.".format(wrapped.__name__)

            # Optionally, add version and alternative.
            if version is not None:
                msg = msg + " It is going to be removed in "
                msg = msg + "the {} version.".format(version)

            if explanation is not None:
                msg = msg + " " + explanation

            # Display the deprecated warning.
            logging.warning(msg)

        # Call the function.
        return wrapped(*args, **kwargs)

    return wrapper(wrapped)



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.collections.nlp.data import BertJointIntentSlotDataset, BertJointIntentSlotInferDataset
from nemo.collections.nlp.nm.data_layers.text_datalayer import TextDataLayer
from nemo.core import ChannelType, LabelsType, MaskType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['BertJointIntentSlotDataLayer', 'BertJointIntentSlotInferDataLayer']


class BertJointIntentSlotDataLayer(TextDataLayer):
    """
    Creates the data layer to use for the task of joint intent
    and slot classification with pretrained model.

    All the data processing is done in BertJointIntentSlotDataset.

    Args:
        input_file (str):
            data file
        slot_file (str):
            file to slot labels, each line corresponding to
            slot labels for a sentence in input_file. No header.
        pad_label (int): pad value use for slot labels
        tokenizer (TokenizerSpec): text tokenizer.
        max_seq_length (int):
            max sequence length minus 2 for [CLS] and [SEP]
        dataset_type (BertJointIntentSlotDataset):
            the dataset that needs to be converted to DataLayerNM
        shuffle (bool): whether to shuffle data or not. Default: False.
        batch_size: text segments batch size
        ignore_extra_tokens (bool): whether or not to ignore extra tokens
        ignore_start_end (bool)": whether or not to ignore start and end
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        input_ids:
            indices of tokens which constitute batches of text segments
        input_type_ids:
            tensor with 0's and 1's to denote the text segment type
        input_mask:
            bool tensor with 0s in place of tokens to be masked
        loss_mask:
            used to mask and ignore tokens in the loss function
        subtokens_mask:
            used to ignore the outputs of unwanted tokens in
            the inference and evaluation like the start and end tokens
        intents:
            intents labels
        slots:
            slots labels
        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_type_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_mask": NeuralType(('B', 'T'), ChannelType()),
            "loss_mask": NeuralType(('B', 'T'), MaskType()),
            "subtokens_mask": NeuralType(('B', 'T'), ChannelType()),
            "intents": NeuralType(tuple('B'), LabelsType()),
            "slots": NeuralType(('B', 'T'), LabelsType()),
        }

    def __init__(
        self,
        input_file,
        slot_file,
        pad_label,
        tokenizer,
        max_seq_length,
        num_samples=-1,
        shuffle=False,
        batch_size=64,
        ignore_extra_tokens=False,
        ignore_start_end=False,
        do_lower_case=False,
        dataset_type=BertJointIntentSlotDataset,
    ):
        dataset_params = {
            'input_file': input_file,
            'slot_file': slot_file,
            'pad_label': pad_label,
            'tokenizer': tokenizer,
            'max_seq_length': max_seq_length,
            'num_samples': num_samples,
            'ignore_extra_tokens': ignore_extra_tokens,
            'ignore_start_end': ignore_start_end,
            'do_lower_case': do_lower_case,
        }
        super().__init__(dataset_type, dataset_params, batch_size, shuffle=shuffle)


class BertJointIntentSlotInferDataLayer(TextDataLayer):
    """
    Creates the data layer to use for the task of joint intent
    and slot classification with pretrained model. This is for

    All the data processing is done in BertJointIntentSlotInferDataset.

    Args:
        queries (list): list of queries for inference
        tokenizer (TokenizerSpec): text tokenizer.
        max_seq_length (int):
            max sequence length minus 2 for [CLS] and [SEP]
        dataset_type (BertJointIntentSlotDataset):
            the dataset that needs to be converted to DataLayerNM
        shuffle (bool): whether to shuffle data or not. Default: False.
        do_lower_case (bool): whether to make the sentence all lower case
        batch_size: text segments batch size
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        
        input_ids:
            indices of tokens which constitute batches of text segments
        input_type_ids:
            tensor with 0's and 1's to denote the text segment type
        input_mask:
            bool tensor with 0s in place of tokens to be masked
        loss_mask:
            used to mask and ignore tokens in the loss function
        subtokens_mask:
            used to ignore the outputs of unwanted tokens in
            the inference and evaluation like the start and end tokens
        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_type_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_mask": NeuralType(('B', 'T'), ChannelType()),
            "loss_mask": NeuralType(('B', 'T'), ChannelType()),
            "subtokens_mask": NeuralType(('B', 'T'), ChannelType()),
        }

    def __init__(
        self,
        queries,
        tokenizer,
        max_seq_length,
        batch_size=1,
        shuffle=False,
        do_lower_case=False,
        dataset_type=BertJointIntentSlotInferDataset,
    ):
        dataset_params = {
            'queries': queries,
            'tokenizer': tokenizer,
            'max_seq_length': max_seq_length,
            'do_lower_case': do_lower_case,
        }
        super().__init__(dataset_type, dataset_params, batch_size, shuffle=shuffle)



# -*- coding: utf-8 -*-

# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from abc import ABC, abstractmethod
from typing import Dict

import nemo
from nemo.core.neural_types import NeuralType


class NeuralInterface(ABC):
    """
        Abstract class offering interface shared between Neural Module and Neural Graph.
        Had to move it to a separate class to:
        a) avoid circular imports between Neural Module and Graph.
        b) avoid collection of init_params implemented by default in Neural Module.
        c) extract only the methods that are shared (NMs have plenty of methods that are not making any sense for
        graph, e.g. get_weights, tie_weights, )
    """

    def __init__(self):
        """
            Constructor. Creates a "shortcut" to the application state.
        """
        # Create access to the app state.
        self._app_state = nemo.utils.app_state.AppState()

    @property
    @abstractmethod
    def input_ports(self) -> Dict[str, NeuralType]:
        """ Returns definitions of module input ports

        Returns:
          A (dict) of module's input ports names to NeuralTypes mapping
        """

    @property
    @abstractmethod
    def output_ports(self) -> Dict[str, NeuralType]:
        """ Returns definitions of module output ports

        Returns:
          A (dict) of module's output ports names to NeuralTypes mapping
        """

    @abstractmethod
    def __call__(self, **kwargs):
        """
            This method is used during the construction of a graph for neural type compatibility checking.
            Actual implementation lies in Neural Module and Neural Graph classes.

        Returns:
          NmTensor object or tuple of NmTensor objects
        """

    @property
    def name(self):
        """ Returns the object name. """
        return self._name



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright 2019 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from unittest import TestCase

import pytest

from nemo.backends.pytorch.tutorials import MSELoss, RealFunctionDataLayer, TaylorNet
from nemo.core.neural_types import (
    AcousticEncodedRepresentation,
    AudioSignal,
    AxisKind,
    AxisKindAbstract,
    AxisType,
    ChannelType,
    ElementType,
    MelSpectrogramType,
    MFCCSpectrogramType,
    NeuralPortNmTensorMismatchError,
    NeuralType,
    NeuralTypeComparisonResult,
    SpectrogramType,
    VoidType,
)


class NeuralTypeSystemTests(TestCase):
    @pytest.mark.unit
    def test_short_vs_long_version(self):
        long_version = NeuralType(
            axes=(AxisType(AxisKind.Batch, None), AxisType(AxisKind.Dimension, None), AxisType(AxisKind.Time, None)),
            elements_type=AcousticEncodedRepresentation(),
        )
        short_version = NeuralType(('B', 'D', 'T'), AcousticEncodedRepresentation())
        self.assertEqual(long_version.compare(short_version), NeuralTypeComparisonResult.SAME)
        self.assertEqual(short_version.compare(long_version), NeuralTypeComparisonResult.SAME)

    @pytest.mark.unit
    def test_parameterized_type_audio_sampling_frequency(self):
        audio16K = NeuralType(axes=('B', 'T'), elements_type=AudioSignal(16000))
        audio8K = NeuralType(axes=('B', 'T'), elements_type=AudioSignal(8000))
        another16K = NeuralType(axes=('B', 'T'), elements_type=AudioSignal(16000))

        self.assertEqual(audio8K.compare(audio16K), NeuralTypeComparisonResult.SAME_TYPE_INCOMPATIBLE_PARAMS)
        self.assertEqual(audio16K.compare(audio8K), NeuralTypeComparisonResult.SAME_TYPE_INCOMPATIBLE_PARAMS)
        self.assertEqual(another16K.compare(audio16K), NeuralTypeComparisonResult.SAME)
        self.assertEqual(audio16K.compare(another16K), NeuralTypeComparisonResult.SAME)

    @pytest.mark.unit
    def test_transpose_same_1(self):
        type1 = NeuralType(axes=('B', 'T', 'C'))
        type2 = NeuralType(axes=('T', 'B', 'C'))
        self.assertEqual(type1.compare(type2), NeuralTypeComparisonResult.TRANSPOSE_SAME)
        self.assertEqual(type2.compare(type1), NeuralTypeComparisonResult.TRANSPOSE_SAME)

    @pytest.mark.unit
    def test_transpose_same_2(self):
        audio16K = NeuralType(axes=('B', 'T'), elements_type=AudioSignal(16000))
        audio16K_t = NeuralType(axes=('T', 'B'), elements_type=AudioSignal(16000))
        self.assertEqual(audio16K.compare(audio16K_t), NeuralTypeComparisonResult.TRANSPOSE_SAME)

    @pytest.mark.unit
    def test_inheritance_spec_augment_example(self):
        input = NeuralType(('B', 'D', 'T'), SpectrogramType())
        out1 = NeuralType(('B', 'D', 'T'), MelSpectrogramType())
        out2 = NeuralType(('B', 'D', 'T'), MFCCSpectrogramType())
        self.assertEqual(out1.compare(out2), NeuralTypeComparisonResult.INCOMPATIBLE)
        self.assertEqual(out2.compare(out1), NeuralTypeComparisonResult.INCOMPATIBLE)
        self.assertEqual(input.compare(out1), NeuralTypeComparisonResult.GREATER)
        self.assertEqual(input.compare(out2), NeuralTypeComparisonResult.GREATER)
        self.assertEqual(out1.compare(input), NeuralTypeComparisonResult.LESS)
        self.assertEqual(out2.compare(input), NeuralTypeComparisonResult.LESS)

    @pytest.mark.unit
    def test_singletone(self):
        loss_output1 = NeuralType(axes=None)
        loss_output2 = NeuralType(axes=None)
        self.assertEqual(loss_output1.compare(loss_output2), NeuralTypeComparisonResult.SAME)
        self.assertEqual(loss_output2.compare(loss_output1), NeuralTypeComparisonResult.SAME)

    @pytest.mark.unit
    def test_list_of_lists(self):
        T1 = NeuralType(
            axes=(
                AxisType(kind=AxisKind.Batch, size=None, is_list=True),
                AxisType(kind=AxisKind.Time, size=None, is_list=True),
                AxisType(kind=AxisKind.Dimension, size=32, is_list=False),
                AxisType(kind=AxisKind.Dimension, size=128, is_list=False),
                AxisType(kind=AxisKind.Dimension, size=256, is_list=False),
            ),
            elements_type=ChannelType(),
        )
        T2 = NeuralType(
            axes=(
                AxisType(kind=AxisKind.Batch, size=None, is_list=False),
                AxisType(kind=AxisKind.Time, size=None, is_list=False),
                AxisType(kind=AxisKind.Dimension, size=32, is_list=False),
                AxisType(kind=AxisKind.Dimension, size=128, is_list=False),
                AxisType(kind=AxisKind.Dimension, size=256, is_list=False),
            ),
            elements_type=ChannelType(),
        )
        # TODO: should this be incompatible instead???
        self.assertEqual(T1.compare(T2), NeuralTypeComparisonResult.TRANSPOSE_SAME)

    @pytest.mark.unit
    def test_void(self):
        btc_spctr = NeuralType(('B', 'T', 'C'), SpectrogramType())
        btc_spct_bad = NeuralType(('B', 'T'), SpectrogramType())
        btc_void = NeuralType(('B', 'T', 'C'), VoidType())
        self.assertEqual(btc_void.compare(btc_spctr), NeuralTypeComparisonResult.SAME)
        self.assertEqual(btc_spctr.compare(btc_void), NeuralTypeComparisonResult.INCOMPATIBLE)
        self.assertEqual(btc_void.compare(btc_spct_bad), NeuralTypeComparisonResult.INCOMPATIBLE)

    @pytest.mark.unit
    def test_big_void(self):
        big_void_1 = NeuralType(elements_type=VoidType())
        big_void_2 = NeuralType()

        btc_spctr = NeuralType(('B', 'T', 'C'), SpectrogramType())
        btc_spct_bad = NeuralType(('B', 'T'), SpectrogramType())
        t1 = NeuralType(
            axes=(
                AxisType(kind=AxisKind.Batch, size=None, is_list=True),
                AxisType(kind=AxisKind.Time, size=None, is_list=True),
                AxisType(kind=AxisKind.Dimension, size=32, is_list=False),
                AxisType(kind=AxisKind.Dimension, size=128, is_list=False),
                AxisType(kind=AxisKind.Dimension, size=256, is_list=False),
            ),
            elements_type=ChannelType(),
        )
        t2 = NeuralType(
            axes=(
                AxisType(kind=AxisKind.Batch, size=None, is_list=False),
                AxisType(kind=AxisKind.Time, size=None, is_list=False),
                AxisType(kind=AxisKind.Dimension, size=32, is_list=False),
                AxisType(kind=AxisKind.Dimension, size=128, is_list=False),
                AxisType(kind=AxisKind.Dimension, size=256, is_list=False),
            ),
            elements_type=ChannelType(),
        )

        self.assertEqual(big_void_1.compare(btc_spctr), NeuralTypeComparisonResult.SAME)
        self.assertEqual(big_void_1.compare(btc_spct_bad), NeuralTypeComparisonResult.SAME)
        self.assertEqual(big_void_1.compare(t1), NeuralTypeComparisonResult.SAME)
        self.assertEqual(big_void_1.compare(t2), NeuralTypeComparisonResult.SAME)

        self.assertEqual(big_void_2.compare(btc_spctr), NeuralTypeComparisonResult.SAME)
        self.assertEqual(big_void_2.compare(btc_spct_bad), NeuralTypeComparisonResult.SAME)
        self.assertEqual(big_void_2.compare(t1), NeuralTypeComparisonResult.SAME)
        self.assertEqual(big_void_2.compare(t2), NeuralTypeComparisonResult.SAME)

    @pytest.mark.unit
    def test_dag(self):
        data_source = RealFunctionDataLayer(n=10000, batch_size=128)
        trainable_module = TaylorNet(dim=4)
        loss = MSELoss()
        x, y = data_source()
        y_pred = trainable_module(x=x)
        _ = loss(predictions=y_pred, target=y)

        def wrong():
            data_source = RealFunctionDataLayer(n=10000, batch_size=128)
            trainable_module = TaylorNet(dim=4)
            loss = MSELoss()
            x, y = data_source()
            loss_tensor = loss(predictions=x, target=x)
            _ = trainable_module(x=loss_tensor)

        self.assertRaises(NeuralPortNmTensorMismatchError, wrong)

    @pytest.mark.unit
    def test_unspecified_dimensions(self):
        t0 = NeuralType(
            (AxisType(AxisKind.Batch, 64), AxisType(AxisKind.Time, 10), AxisType(AxisKind.Dimension, 128)),
            SpectrogramType(),
        )
        t1 = NeuralType(('B', 'T', 'C'), SpectrogramType())
        self.assertEqual(t1.compare(t0), NeuralTypeComparisonResult.SAME)
        self.assertEqual(t0.compare(t1), NeuralTypeComparisonResult.DIM_INCOMPATIBLE)

    @pytest.mark.unit
    def test_any_axis(self):
        t0 = NeuralType(('B', 'Any', 'Any'), VoidType())
        t1 = NeuralType(('B', 'Any', 'Any'), SpectrogramType())
        t2 = NeuralType(('B', 'T', 'C'), SpectrogramType())
        self.assertEqual(t0.compare(t1), NeuralTypeComparisonResult.SAME)
        self.assertEqual(t0.compare(t2), NeuralTypeComparisonResult.SAME)
        self.assertEqual(t1.compare(t2), NeuralTypeComparisonResult.SAME)
        self.assertEqual(t2.compare(t1), NeuralTypeComparisonResult.INCOMPATIBLE)
        self.assertEqual(t1.compare(t0), NeuralTypeComparisonResult.INCOMPATIBLE)

    @pytest.mark.unit
    def test_struct(self):
        class BoundingBox(ElementType):
            def __str__(self):
                return "bounding box from detection model"

            def fields(self):
                return ("X", "Y", "W", "H")

        # ALSO ADD new, user-defined, axis kind
        class AxisKind2(AxisKindAbstract):
            Image = 0

        T1 = NeuralType(
            elements_type=BoundingBox(),
            axes=(
                AxisType(kind=AxisKind.Batch, size=None, is_list=True),
                AxisType(kind=AxisKind2.Image, size=None, is_list=True),
            ),
        )

        class BadBoundingBox(ElementType):
            def __str__(self):
                return "bad bounding box from detection model"

            def fields(self):
                return ("X", "Y", "H")

        T2 = NeuralType(
            elements_type=BadBoundingBox(),
            axes=(
                AxisType(kind=AxisKind.Batch, size=None, is_list=True),
                AxisType(kind=AxisKind2.Image, size=None, is_list=True),
            ),
        )
        self.assertEqual(T2.compare(T1), NeuralTypeComparisonResult.INCOMPATIBLE)



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import torch
import torch.nn as nn

__all__ = ['form_attention_mask', 'transformer_weights_init']

NEG_INF = -10000.0


def form_attention_mask(input_mask, diagonal=None):
    """
    Build attention mask with optional masking of future tokens we forbid
    to attend to (e.g. as it is in Transformer decoder).

    Args:
        input_mask: binary mask of size B x L with 1s corresponding to valid
            tokens and 0s corresponding to padding tokens
        diagonal: diagonal where triangular future mask starts
            None -- do not mask anything
            0 -- regular translation or language modeling future masking
            1 -- query stream masking as in XLNet architecture
    Returns:
        attention_mask: mask of size B x 1 x L x L with 0s corresponding to
            tokens we plan to attend to and -10000 otherwise
    """

    if input_mask is None:
        return None
    attn_shape = (1, input_mask.shape[1], input_mask.shape[1])
    attn_mask = input_mask.byte().unsqueeze(1)
    if diagonal is not None:
        future_mask = torch.tril(torch.ones(attn_shape).byte().to(input_mask.device), diagonal)
        attn_mask = attn_mask & future_mask
    attention_mask = (1 - attn_mask.to(torch.float)) * NEG_INF
    return attention_mask.unsqueeze(1)


def transformer_weights_init(module, std_init_range=0.02, xavier=True):
    """
    Initialize different weights in Transformer model.

    Args:
        module: torch.nn.Module to be initialized
        std_init_range: standard deviation of normal initializer
        xavier: if True, xavier initializer will be used in Linear layers
            as was proposed in AIAYN paper, otherwise normal initializer
            will be used (like in BERT paper)
    """

    if isinstance(module, nn.Linear):
        if xavier:
            nn.init.xavier_uniform_(module.weight)
        else:
            nn.init.normal_(module.weight, mean=0.0, std=std_init_range)
        if module.bias is not None:
            nn.init.constant_(module.bias, 0.0)
    elif isinstance(module, nn.Embedding):
        nn.init.normal_(module.weight, mean=0.0, std=std_init_range)
    elif isinstance(module, nn.LayerNorm):
        nn.init.constant_(module.weight, 1.0)
        nn.init.constant_(module.bias, 0.0)



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""Pytorch Dataset for training BERT."""

import array
import glob
import os
import pickle
import random

import h5py
import numpy as np
from sentencepiece import SentencePieceTrainer as SPT
from torch.utils.data import Dataset
from tqdm import tqdm

from nemo import logging
from nemo.collections.nlp.data.datasets.datasets_utils.data_preprocessing import DATABASE_EXISTS_TMP, if_exist

__all__ = ['BertPretrainingDataset', 'BertPretrainingPreprocessedDataset']


class BertPretrainingDataset(Dataset):
    def __init__(
        self,
        tokenizer,
        dataset,
        max_seq_length=128,
        mask_probability=0.15,
        short_seq_prob=0.1,
        seq_a_ratio=0.6,
        sentence_idx_file=None,
    ):
        self.tokenizer = tokenizer

        # Loading enormous datasets into RAM isn't always feasible -- for
        # example, the pubmed corpus is 200+ GB, which doesn't fit into RAM on
        # most computers. To get around this, we store the indices of newlines
        # in each file so we can seek to and retrieve sentences immediately
        # from main memory when needed during training.

        if sentence_idx_file is None:
            data_dir = dataset[: dataset.rfind('/')]
            mode = dataset[dataset.rfind('/') + 1 : dataset.rfind('.')]
            sentence_idx_file = f"{data_dir}/{mode}_sentence_indices.pkl"

        if os.path.isfile(sentence_idx_file):
            # If the sentence indices file already exists, load from it
            with open(sentence_idx_file, "rb") as f:
                sentence_indices = pickle.load(f)
        else:
            # Otherwise, generate and store sentence indices
            sentence_indices = {}

            # Finds all of the newline indices in a string
            def find_newlines(contents):
                start = 0

                while True:
                    try:
                        # index and split are much faster than Python for loops
                        new_start = contents.index(b"\n", start)
                        line = (
                            contents[start:new_start]
                            .replace(b"\xc2\x99", b" ")
                            .replace(b"\xc2\xa0", b" ")
                            .decode("utf-8", errors="ignore")
                        )

                        if len(line.split()) > 0:
                            yield start

                        start = new_start + 1

                    except ValueError:
                        break

            if os.path.isdir(dataset):
                dataset_pattern = os.path.join(dataset, "**", "*.txt")
                filenames = glob.glob(dataset_pattern, recursive=True)
            else:
                filenames = [dataset]

            for filename in tqdm(filenames):
                with open(filename, "rb") as f:
                    contents = f.read()
                    newline_indices = find_newlines(contents)

                if os.path.isdir(dataset):
                    # Only keep the parts of the filepath that are invariant to
                    # the dataset's location on disk
                    filename = os.path.basename(filename)

                # In python, arrays are much more space-efficient than lists
                sentence_indices[filename] = array.array("I", newline_indices)

            # Save sentence indices so we don't have to do this again
            with open(sentence_idx_file, "wb") as f:
                pickle.dump(sentence_indices, f)

        corpus_size = 0
        empty_files = []

        # Find total number of newlines across entire corpus and remove files
        # without any newlines
        for filename in sentence_indices:
            if len(sentence_indices[filename]) <= 1:
                empty_files.append(filename)
            else:
                corpus_size += len(sentence_indices[filename])

        for filename in empty_files:
            del sentence_indices[filename]

        self.corpus_size = corpus_size
        self.dataset = dataset
        self.filenames = list(sentence_indices.keys())
        self.mask_probability = mask_probability
        self.max_seq_length = max_seq_length
        self.sentence_indices = sentence_indices
        self.vocab_size = self.tokenizer.vocab_size
        self.short_seq_prob = short_seq_prob
        self.seq_a_ratio = seq_a_ratio

    def __len__(self):
        return self.corpus_size

    def __getitem__(self, idx, min_doc_length=16):
        # Each sequence has three special tokens, as follows:
        # tokenizer.cls_token <document a> tokenizer.sep_token <document b> tokenizer.eos_token
        num_special_tokens = 3

        max_num_tokens = self.max_seq_length - num_special_tokens
        target_seq_length = max_num_tokens
        if random.random() < self.short_seq_prob:
            # TODO: maybe introduce an argument to control this.
            target_seq_length = random.randint(2, max_num_tokens)

        # prefer the seq_a to be slightly longer than seq_b, 0.6 by default
        target_seq_length_a = int(round(target_seq_length * self.seq_a_ratio))
        target_seq_length_b = target_seq_length - target_seq_length_a

        def get_document(filepath, offset):
            # Retrieve a specific line from a file and return as a document
            if os.path.isdir(self.dataset):
                filepath = os.path.join(self.dataset, filepath)

            with open(filepath, "rb") as f:
                f.seek(offset)
                doc_text = f.readline()[:-1].decode("utf-8", errors="ignore")
                document = self.tokenizer.text_to_ids(doc_text)

            return document

        def match_target_seq_length(document, target_seq_length, filename, line_idx, sentence_indices):
            # If document is shorter than target sequence length,
            # append the next line or take a random line as replacement.
            num_lines = len(sentence_indices[filename])

            while len(document) < target_seq_length:
                if line_idx < (num_lines - 1):
                    # append the next line
                    line_idx += 1
                else:
                    # current line is the last line, take a random one
                    line_idx = random.randrange(num_lines)
                    document = []

                offset = sentence_indices[filename][line_idx]
                document += get_document(filename, offset)

            return document, line_idx

        # Take sequence A from a random file and a random line
        a_filename = random.choice(self.filenames)
        a_line_idx = random.randrange(len(self.sentence_indices[a_filename]))
        a_line_offset = self.sentence_indices[a_filename][a_line_idx]
        a_document = get_document(a_filename, a_line_offset)
        a_document, a_line_idx = match_target_seq_length(
            a_document, target_seq_length_a, a_filename, a_line_idx, self.sentence_indices
        )

        is_last_line = a_line_idx >= (len(self.sentence_indices[a_filename]) - 1)
        # About 50% of the time, B is a random sentence from the corpus
        take_random_b = (random.random() < 0.5) or is_last_line

        if take_random_b:
            # This should rarely go for more than one iteration for large
            # corpora. However, just to be careful, we try to make sure that
            # the random document is not the same as the document
            # we're processing.
            for _ in range(10):
                b_filename = random.choice(self.filenames)
                b_line_idx = random.choice(range(len(self.sentence_indices[b_filename])))
                if b_filename != a_filename:
                    break
                else:
                    # Take another line from the same file
                    b_line_pos = self.sentence_indices[b_filename][b_line_idx]
                    a_line_pos = self.sentence_indices[a_filename][a_line_idx]
                    # TODO unclear about the following check
                    if abs(b_line_pos - a_line_pos) > max_num_tokens:
                        break
                    else:
                        pass
        else:
            b_filename = a_filename
            b_line_idx = a_line_idx + 1

        is_next = int(not take_random_b)
        b_line_pos = self.sentence_indices[b_filename][b_line_idx]
        b_document = get_document(b_filename, b_line_pos)
        b_document, b_line_idx = match_target_seq_length(
            b_document, target_seq_length_b, b_filename, b_line_idx, self.sentence_indices
        )

        def truncate_seq_pair(a, b, max_num_tokens):
            # Truncates a pair of sequences to a maximum sequence length
            while (len(a) + len(b)) > max_num_tokens:
                # Truncate the longer sequence
                if len(a) > len(b):
                    trunc_document = a
                else:
                    trunc_document = b

                if len(trunc_document) <= 1:
                    raise ValueError(
                        "Input text corpora probably too small. "
                        "Failed to truncate sequence pair to "
                        "maximum sequence legnth."
                    )

                # Randomly truncate from the front or the back
                if random.random() < 0.5:
                    del trunc_document[0]
                else:
                    trunc_document.pop()

        truncate_seq_pair(a_document, b_document, max_num_tokens)

        output_ids = (
            [self.tokenizer.cls_id] + a_document + [self.tokenizer.sep_id] + b_document + [self.tokenizer.eos_id]
        )

        input_ids, output_mask = self.mask_ids(output_ids)

        input_mask = np.zeros(self.max_seq_length, dtype=np.long)
        input_mask[: len(input_ids)] = 1

        input_type_ids = np.zeros(self.max_seq_length, dtype=np.int)
        input_type_ids[len(a_document) + 2 : len(output_ids) + 1] = 1

        padding_length = max(0, self.max_seq_length - len(input_ids))
        if padding_length > 0:
            input_ids.extend([self.tokenizer.pad_id] * padding_length)
            output_ids.extend([self.tokenizer.pad_id] * padding_length)
            output_mask.extend([0] * padding_length)

        # TODO: wrap the return value with () for consistent style.
        return (
            np.array(input_ids),
            input_type_ids,
            np.array(input_mask, dtype=np.long),
            np.array(output_ids),
            np.array(output_mask, dtype=np.float32),
            is_next,
        )

    def mask_ids(self, ids):
        """
        Args:
          ids: list of token ids representing a chunk of text
        Returns:
          masked_ids: list of input tokens with some of the entries masked
            according to the following protocol from the original BERT paper:
            each token is masked with a probability of 15% and is replaced with
            1) the [MASK] token 80% of the time,
            2) random token 10% of the time,
            3) the same token 10% of the time.
          output_mask: list of binary variables which indicate what tokens has
            been masked (to calculate the loss function for these tokens only)
        """

        # Whole-word masking by default, as it gives better performance.
        cand_indexes = [[ids[0]]]
        for tid in ids[1:]:
            token = self.tokenizer.ids_to_tokens([tid])[0]
            is_suffix = token.startswith('\u2581')
            if is_suffix:
                # group together with its previous token to form a whole-word
                cand_indexes[-1].append(tid)
            else:
                cand_indexes.append([tid])

        masked_ids, output_mask = [], []
        mask_id = self.tokenizer.token_to_id("[MASK]")

        for word_ids in cand_indexes:
            is_special = (word_ids[0] == self.tokenizer.cls_id) or (word_ids[0] == self.tokenizer.sep_id)
            if is_special or (random.random() > self.mask_probability):
                output_mask.extend([0] * len(word_ids))
                masked_ids.extend(word_ids)
            else:
                output_mask.extend([1] * len(word_ids))
                p = random.random()
                # for 80%, replace with mask
                if p < 0.8:
                    masked_ids.extend([mask_id] * len(word_ids))
                # for 10%, replace by a random token
                elif p < 0.9:
                    for _ in word_ids:
                        # randomly select a valid word
                        random_word = random.randrange(self.vocab_size)
                        while random_word in (self.tokenizer.cls_id, self.tokenizer.sep_id):
                            random_word = random.randrange(self.vocab_size)
                        masked_ids.append(random_word)
                # for 10%, use same token
                else:
                    masked_ids.extend(word_ids)

        return masked_ids, output_mask


class BertPretrainingPreprocessedDataset(Dataset):
    def __init__(self, input_file, max_pred_length):
        self.input_file = input_file
        self.max_pred_length = max_pred_length
        f = h5py.File(input_file, "r")
        keys = [
            'input_ids',
            'input_mask',
            'segment_ids',
            'masked_lm_positions',
            'masked_lm_ids',
            'next_sentence_labels',
        ]
        self.inputs = [np.asarray(f[key][:]) for key in keys]
        f.close()

    def __len__(self):
        'Denotes the total number of samples'
        return len(self.inputs[0])

    def __getitem__(self, index):
        [input_ids, input_mask, segment_ids, masked_lm_positions, masked_lm_ids, next_sentence_labels] = [
            input[index].astype(np.int64) for input in self.inputs
        ]

        output_mask = np.zeros_like(input_ids)
        output_ids = input_ids.copy()

        index = self.max_pred_length
        padded_mask_indices = (masked_lm_positions == 0).nonzero()
        if len(padded_mask_indices[0]) != 0:
            index = padded_mask_indices[0][0]

        output_mask[masked_lm_positions[:index]] = 1.0
        output_ids[masked_lm_positions[:index]] = masked_lm_ids[:index]

        input_mask = np.asarray(input_mask, dtype=np.float32)
        output_mask = np.asarray(output_mask, dtype=np.float32)
        return (input_ids, segment_ids, input_mask, output_ids, output_mask, next_sentence_labels)


class BERTPretrainingDataDesc:
    def __init__(
        self, dataset_name, vocab_size, sample_size, special_tokens, train_data, eval_data=None, test_data=None,
    ):
        if dataset_name == 'wikitext-2':
            if not os.path.exists(train_data):
                raise FileNotFoundError(
                    "Dataset not found. Run 'get_wkt2.sh DATA_DIR' from examples/nlp/language_modeling"
                )
            self.data_dir, self.tokenizer_model, self.vocab_file = self.create_vocab_mlm(
                train_data, vocab_size, sample_size, special_tokens
            )
        else:
            raise ValueError(
                "Looks like you passed a dataset name that isn't already supported by NeMo. Please make sure that "
                "you build the preprocessing method for it."
            )

        self.train_file = train_data
        self.eval_file = eval_data
        self.test_file = test_data

    def create_vocab_mlm(
        self, data_file, vocab_size, sample_size, special_tokens=['[PAD]', '[UNK]', '[CLS]', '[SEP]', '[MASK]'],
    ):
        data_dir = os.path.dirname(data_file)
        vocab = special_tokens[:]
        bert_dir = f'{data_dir}/bert'
        if if_exist(bert_dir, ['tokenizer.model']):
            logging.info(DATABASE_EXISTS_TMP.format('WikiText_BERT', bert_dir))
            return data_dir, f'{bert_dir}/tokenizer.model', f'{bert_dir}/vocab.txt'
        logging.info(f'Processing WikiText dataset and store at {bert_dir}')
        os.makedirs(bert_dir, exist_ok=True)

        if not data_file:
            files = glob.glob(f'{data_dir}/*.txt')
            data_file = f'{bert_dir}/merged.txt'
            logging.info(f"Merging {len(files)} txt files into {data_file}")

            with open(data_file, "w") as merged:
                for file in tqdm(files):
                    with open(file, 'r') as inf:
                        content = inf.read().strip()
                    merged.write(content + '\n\n\n')

        cmd = (
            f"--input={data_file} --model_prefix={bert_dir}/tokenizer "
            f"--vocab_size={vocab_size - len(vocab)} "
            f"--input_sentence_size={sample_size} "
            f"--shuffle_input_sentence=true --hard_vocab_limit=false "
            f"--bos_id=-1 --eos_id=-1"
        )

        SPT.Train(cmd)

        # Add BERT control symbols
        tokens = []

        with open(f"{bert_dir}/tokenizer.vocab", "r") as f:
            f.readline()  # skip first <unk> token

            # Read tokens from each line and parse for vocab
            for line in f:
                piece = line.split("\t")[0]
                token = piece[1:] if piece.startswith("▁") else f"##{piece}"
                tokens.append(token)

        vocab.extend(tokens)

        # Save vocabulary to output file
        vocab_file = f'{bert_dir}/vocab.txt'
        with open(vocab_file, "w") as f:
            for token in vocab:
                f.write(f"{token}\n".format())
        return data_dir, f'{bert_dir}/tokenizer.model', vocab_file



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

'''
This file contains code artifacts adapted from the original implementation:
https://github.com/google-research/google-research/blob/master/schema_guided_dst/baseline/train_and_predict.py
'''

import math

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from nemo.backends.pytorch.nm import TrainableNM
from nemo.core import ChannelType, EmbeddedTextType, LogitsType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['SGDDecoderNM']


class LogitsAttention(nn.Module):
    def __init__(self, num_classes, embedding_dim):
        """Get logits for elements by using attention on token embedding.
        Args:
          num_classes (int): An int containing the number of classes for which logits are to be generated.
          embedding_dim (int): hidden size of the BERT
    
        Returns:
          A tensor of shape (batch_size, num_elements, num_classes) containing the logits.
        """
        super().__init__()
        self.num_attention_heads = 16
        self.attention_head_size = embedding_dim // self.num_attention_heads
        self.embedding_dim = embedding_dim
        self.num_classes = num_classes
        self.dropout = nn.Dropout(0.1)

        self.key = nn.Linear(embedding_dim, embedding_dim)
        self.query = nn.Linear(embedding_dim, embedding_dim)
        self.value = nn.Linear(embedding_dim, embedding_dim)
        self.layer = nn.Linear(embedding_dim, num_classes)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(self, encoded_utterance, token_embeddings, element_embeddings, utterance_mask):
        """
        token_embeddings: token hidden states from BERT encoding of the utterance
        encoded_utterance: [CLS] token hidden state from BERT encoding of the utterance
        element_embeddings: A tensor of shape (batch_size, num_elements, embedding_dim) extracted from schema
        utterance_mask: binary mask for token_embeddings, 1 for real tokens 0 for padded tokens
        """
        _, num_elements, _ = element_embeddings.size()

        query_layer = self.query(element_embeddings)
        key_layer = self.key(token_embeddings)
        value_layer = self.value(token_embeddings)

        query_layer = self.transpose_for_scores(query_layer)
        key_layer = self.transpose_for_scores(key_layer)
        value_layer = self.transpose_for_scores(value_layer)
        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))
        attention_scores = attention_scores / math.sqrt(self.attention_head_size)
        if utterance_mask is not None:
            negative_scores = (torch.finfo(attention_scores.dtype).max * -0.7) * torch.ones_like(attention_scores)
            new_x_shape = (utterance_mask.size()[0],) + (1, 1) + (utterance_mask.size()[1],)
            attention_scores = torch.where(
                utterance_mask.view(*new_x_shape).to(bool), attention_scores, negative_scores
            )

        attention_probs = nn.Softmax(dim=-1)(attention_scores)
        attention_probs = self.dropout(attention_probs)
        context_layer = torch.matmul(attention_probs, value_layer)
        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.embedding_dim,)
        context_layer = context_layer.view(*new_context_layer_shape)

        logits = self.layer(context_layer)
        return logits


class Logits(nn.Module):
    def __init__(self, num_classes, embedding_dim):
        """Get logits for elements by conditioning on utterance embedding.
        Args:
          num_classes (int): An int containing the number of classes for which logits are to be generated.
          embedding_dim (int): hidden size of the BERT
    
        Returns:
          A tensor of shape (batch_size, num_elements, num_classes) containing the logits.
        """
        super().__init__()
        self.num_classes = num_classes
        self.utterance_proj = nn.Linear(embedding_dim, embedding_dim)
        self.activation = F.gelu

        self.layer1 = nn.Linear(2 * embedding_dim, embedding_dim)
        self.layer2 = nn.Linear(embedding_dim, num_classes)

    def forward(self, encoded_utterance, token_embeddings, element_embeddings, utterance_mask):
        """
        token_embeddings - token hidden states from BERT encoding of the utterance. Not used
        encoded_utterance - [CLS] token hidden state from BERT encoding of the utterance
        element_embeddings: A tensor of shape (batch_size, num_elements, embedding_dim).
        utterance_mask: binary mask for token_embeddings, 1 for real tokens 0 for padded tokens. Not used
        """
        _, num_elements, _ = element_embeddings.size()

        # Project the utterance embeddings.
        utterance_embedding = self.utterance_proj(encoded_utterance)
        utterance_embedding = self.activation(utterance_embedding)

        # Combine the utterance and element embeddings.
        repeated_utterance_embedding = utterance_embedding.unsqueeze(1).repeat(1, num_elements, 1)

        utterance_element_emb = torch.cat([repeated_utterance_embedding, element_embeddings], axis=2)
        logits = self.layer1(utterance_element_emb)
        logits = self.activation(logits)
        logits = self.layer2(logits)
        return logits


class SGDDecoderNM(TrainableNM):
    """
    Baseline model for schema guided dialogue state tracking with option to make schema embeddings learnable
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module output ports.
        encoded_utterance (float): [CLS] token hidden state from BERT encoding of the utterance
        token_embeddings (float): BERT encoding of utterance (all tokens)
        utterance_mask (bool): Mask which takes the value 0 for padded tokens and 1 otherwise
        cat_slot_values_mask (int): Masks out categorical slots values for slots not used in the service, takes values 0 and 1
        intent_status_mask (int): Masks out padded intents in the service, takes values 0 and 1
        service_ids (int): service ids
        """
        return {
            "encoded_utterance": NeuralType(('B', 'T'), EmbeddedTextType()),
            "token_embeddings": NeuralType(('B', 'T', 'C'), ChannelType()),
            "utterance_mask": NeuralType(('B', 'T'), ChannelType()),
            "cat_slot_values_mask": NeuralType(('B', 'T', 'C'), ChannelType()),
            "intent_status_mask": NeuralType(('B', 'T'), ChannelType()),
            "service_ids": NeuralType(('B'), ChannelType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
            logit_intent_status (float): output for intent status
            logit_req_slot_status (float): output for requested slots status
            logit_cat_slot_status (float): output for categorical slots status
            logit_cat_slot_value (float): output for categorical slots values
            logit_noncat_slot_status (float): Output of SGD model
            logit_noncat_slot_start (float): output for non categorical slots values start
            logit_noncat_slot_end (float): output for non categorical slots values end
        """
        return {
            "logit_intent_status": NeuralType(('B', 'T', 'C'), LogitsType()),
            "logit_req_slot_status": NeuralType(('B', 'T'), LogitsType()),
            "logit_cat_slot_status": NeuralType(('B', 'T', 'C'), LogitsType()),
            "logit_cat_slot_value": NeuralType(('B', 'T', 'C'), LogitsType()),
            "logit_noncat_slot_status": NeuralType(('B', 'T', 'C'), LogitsType()),
            "logit_noncat_slot_start": NeuralType(('B', 'T', 'C'), LogitsType()),
            "logit_noncat_slot_end": NeuralType(('B', 'T', 'C'), LogitsType()),
        }

    def __init__(self, embedding_dim, schema_emb_processor, add_attention_head=False):
        """Get logits for elements by conditioning on utterance embedding.

        Args:
            embedding_dim (int): hidden size of the BERT
            schema_emb_processor (obj): contains schema embeddings for services and config file
            head_transform (str): transformation to use for computing head
        """
        super().__init__()

        # Add a trainable vector for the NONE intent
        self.none_intent_vector = torch.empty((1, 1, embedding_dim), requires_grad=True).to(self._device)
        # TODO truncated norm init
        nn.init.normal_(self.none_intent_vector, std=0.02)
        self.none_intent_vector = torch.nn.Parameter(self.none_intent_vector).to(self._device)

        if add_attention_head:
            projection_module = LogitsAttention
        else:
            projection_module = Logits

        self.intent_layer = Logits(1, embedding_dim).to(self._device)
        self.requested_slots_layer = Logits(1, embedding_dim).to(self._device)

        self.cat_slot_value_layer = projection_module(1, embedding_dim).to(self._device)

        # Slot status values: none, dontcare, active.
        self.cat_slot_status_layer = projection_module(schema_emb_processor._slot_status_size, embedding_dim).to(
            self._device
        )
        self.noncat_slot_layer = projection_module(schema_emb_processor._slot_status_size, embedding_dim).to(
            self._device
        )

        # dim 2 for non_categorical slot - to represent start and end position
        self.noncat_layer1 = nn.Linear(2 * embedding_dim, embedding_dim).to(self._device)
        self.noncat_activation = F.gelu
        self.noncat_layer2 = nn.Linear(embedding_dim, 2).to(self._device)

        config = schema_emb_processor.schema_config
        num_services = len(schema_emb_processor.schemas.services)
        self.intents_emb = nn.Embedding(num_services, config["MAX_NUM_INTENT"] * embedding_dim)
        self.cat_slot_emb = nn.Embedding(num_services, config["MAX_NUM_CAT_SLOT"] * embedding_dim)
        self.cat_slot_value_emb = nn.Embedding(
            num_services, config["MAX_NUM_CAT_SLOT"] * config["MAX_NUM_VALUE_PER_CAT_SLOT"] * embedding_dim
        )
        self.noncat_slot_emb = nn.Embedding(num_services, config["MAX_NUM_NONCAT_SLOT"] * embedding_dim)
        self.req_slot_emb = nn.Embedding(
            num_services, (config["MAX_NUM_CAT_SLOT"] + config["MAX_NUM_NONCAT_SLOT"]) * embedding_dim
        )

        # initialize schema embeddings from the BERT generated embeddings
        schema_embeddings = schema_emb_processor.get_schema_embeddings()
        self.intents_emb.weight.data.copy_(
            torch.from_numpy(np.stack(schema_embeddings['intent_emb']).reshape(num_services, -1))
        )
        self.cat_slot_emb.weight.data.copy_(
            torch.from_numpy(np.stack(schema_embeddings['cat_slot_emb']).reshape(num_services, -1))
        )
        self.cat_slot_value_emb.weight.data.copy_(
            torch.from_numpy(np.stack(schema_embeddings['cat_slot_value_emb']).reshape(num_services, -1))
        )
        self.noncat_slot_emb.weight.data.copy_(
            torch.from_numpy(np.stack(schema_embeddings['noncat_slot_emb']).reshape(num_services, -1))
        )
        self.req_slot_emb.weight.data.copy_(
            torch.from_numpy(np.stack(schema_embeddings['req_slot_emb']).reshape(num_services, -1))
        )

        if not schema_emb_processor.is_trainable:
            self.intents_emb.weight.requires_grad = False
            self.cat_slot_emb.weight.requires_grad = False
            self.cat_slot_value_emb.weight.requires_grad = False
            self.noncat_slot_emb.weight.requires_grad = False
            self.req_slot_emb.weight.requires_grad = False

        self.to(self._device)

    def forward(
        self,
        encoded_utterance,
        token_embeddings,
        utterance_mask,
        cat_slot_values_mask,
        service_ids,
        intent_status_mask,
    ):
        batch_size, emb_dim = encoded_utterance.size()
        intent_embeddings = self.intents_emb(service_ids).view(batch_size, -1, emb_dim)
        cat_slot_emb = self.cat_slot_emb(service_ids).view(batch_size, -1, emb_dim)
        max_number_cat_slots = cat_slot_emb.shape[1]
        cat_slot_value_emb = self.cat_slot_value_emb(service_ids).view(batch_size, max_number_cat_slots, -1, emb_dim)
        noncat_slot_emb = self.noncat_slot_emb(service_ids).view(batch_size, -1, emb_dim)
        req_slot_emb = self.req_slot_emb(service_ids).view(batch_size, -1, emb_dim)

        logit_intent_status = self._get_intents(
            encoded_utterance, intent_embeddings, intent_status_mask, token_embeddings, utterance_mask
        )

        logit_req_slot_status = self._get_requested_slots(
            encoded_utterance, req_slot_emb, token_embeddings, utterance_mask
        )

        logit_cat_slot_status, logit_cat_slot_value = self._get_categorical_slot_goals(
            encoded_utterance, cat_slot_emb, cat_slot_value_emb, cat_slot_values_mask, token_embeddings, utterance_mask
        )

        (
            logit_noncat_slot_status,
            logit_noncat_slot_start,
            logit_noncat_slot_end,
        ) = self._get_noncategorical_slot_goals(encoded_utterance, utterance_mask, noncat_slot_emb, token_embeddings)

        return (
            logit_intent_status,
            logit_req_slot_status,
            logit_cat_slot_status,
            logit_cat_slot_value,
            logit_noncat_slot_status,
            logit_noncat_slot_start,
            logit_noncat_slot_end,
        )

    def _get_intents(self, encoded_utterance, intent_embeddings, intent_status_mask, token_embeddings, utterance_mask):
        """
        Args:
            intent_embedding - BERT schema embeddings
            encoded_utterance - representation of untterance
            intent_status_mask - masks out intent not used for the service
        """
        batch_size = intent_embeddings.size()[0]

        # Add a trainable vector for the NONE intent.
        repeated_none_intent_vector = self.none_intent_vector.repeat(batch_size, 1, 1)
        intent_embeddings = torch.cat([repeated_none_intent_vector, intent_embeddings], axis=1)
        logits = self.intent_layer(
            encoded_utterance=encoded_utterance,
            token_embeddings=token_embeddings,
            element_embeddings=intent_embeddings,
            utterance_mask=utterance_mask,
        )
        logits = logits.squeeze(axis=-1)  # Shape: (batch_size, max_intents + 1)

        # Mask out logits for padded intents
        negative_logits = self._get_negative_logits(logits)
        return torch.where(intent_status_mask.to(dtype=torch.bool), logits, negative_logits)

    def _get_requested_slots(self, encoded_utterance, requested_slot_emb, token_embeddings, utterance_mask):
        """Obtain logits for requested slots."""

        logits = self.requested_slots_layer(
            encoded_utterance=encoded_utterance,
            token_embeddings=token_embeddings,
            element_embeddings=requested_slot_emb,
            utterance_mask=utterance_mask,
        )
        logits = logits.squeeze(axis=-1)

        # logits shape: (batch_size, max_num_slots)
        logits = logits.squeeze(axis=-1)
        return logits

    def _get_categorical_slot_goals(
        self,
        encoded_utterance,
        cat_slot_emb,
        cat_slot_value_emb,
        cat_slot_values_mask,
        token_embeddings,
        utterance_mask,
    ):
        """
        Obtain logits for status and values for categorical slots
        Slot status values: none, dontcare, active
        """

        # Predict the status of all categorical slots.
        status_logits = self.cat_slot_status_layer(
            encoded_utterance=encoded_utterance,
            token_embeddings=token_embeddings,
            element_embeddings=cat_slot_emb,
            utterance_mask=utterance_mask,
        )

        # Predict the goal value.
        # Shape: (batch_size, max_categorical_slots, max_categorical_values, embedding_dim).
        _, max_num_slots, max_num_values, embedding_dim = cat_slot_value_emb.size()
        cat_slot_value_emb_reshaped = cat_slot_value_emb.view(-1, max_num_slots * max_num_values, embedding_dim)

        value_logits = self.cat_slot_value_layer(
            encoded_utterance=encoded_utterance,
            token_embeddings=token_embeddings,
            element_embeddings=cat_slot_value_emb_reshaped,
            utterance_mask=utterance_mask,
        )

        # Reshape to obtain the logits for all slots.
        value_logits = value_logits.view(-1, max_num_slots, max_num_values)

        # Mask out logits for padded slots and values because they will be softmaxed
        negative_value_logits = self._get_negative_logits(value_logits)
        value_logits = torch.where(cat_slot_values_mask.to(dtype=torch.bool), value_logits, negative_value_logits)
        return status_logits, value_logits

    def _get_noncategorical_slot_goals(self, encoded_utterance, utterance_mask, noncat_slot_emb, token_embeddings):
        """
        Obtain logits for status and slot spans for non-categorical slots.
        Slot status values: none, dontcare, active
        """
        # Predict the status of all non-categorical slots.
        max_num_slots = noncat_slot_emb.size()[1]
        status_logits = self.noncat_slot_layer(
            encoded_utterance=encoded_utterance,
            token_embeddings=token_embeddings,
            element_embeddings=noncat_slot_emb,
            utterance_mask=utterance_mask,
        )

        # Predict the distribution for span indices.
        max_num_tokens = token_embeddings.size()[1]

        repeated_token_embeddings = token_embeddings.unsqueeze(1).repeat(1, max_num_slots, 1, 1)
        repeated_slot_embeddings = noncat_slot_emb.unsqueeze(2).repeat(1, 1, max_num_tokens, 1)

        # Shape: (batch_size, max_num_slots, max_num_tokens, 2 * embedding_dim).
        slot_token_embeddings = torch.cat([repeated_slot_embeddings, repeated_token_embeddings], axis=3)

        # Project the combined embeddings to obtain logits, Shape: (batch_size, max_num_slots, max_num_tokens, 2)
        span_logits = self.noncat_layer1(slot_token_embeddings)
        span_logits = self.noncat_activation(span_logits)
        span_logits = self.noncat_layer2(span_logits)

        # Mask out invalid logits for padded tokens.
        utterance_mask = utterance_mask.to(bool)  # Shape: (batch_size, max_num_tokens).
        repeated_utterance_mask = utterance_mask.unsqueeze(1).unsqueeze(3).repeat(1, max_num_slots, 1, 2)
        negative_logits = (torch.finfo(span_logits.dtype).max * -0.7) * torch.ones(
            span_logits.size(), device=self._device, dtype=span_logits.dtype
        )

        span_logits = torch.where(repeated_utterance_mask, span_logits, negative_logits)

        # Shape of both tensors: (batch_size, max_num_slots, max_num_tokens).
        span_start_logits, span_end_logits = torch.unbind(span_logits, dim=3)
        return status_logits, span_start_logits, span_end_logits

    def _get_negative_logits(self, logits):
        # returns tensor with negative logits that will be used to mask out unused values
        # for a particular service
        negative_logits = (torch.finfo(logits.dtype).max * -0.7) * torch.ones(
            logits.size(), device=self._device, dtype=logits.dtype
        )
        return negative_logits



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2018 The Google AI Language Team Authors and
# The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import collections

from transformers.tokenization_bert import BasicTokenizer

from nemo import logging
from nemo.collections.nlp.utils.data_utils import get_tokens, normalize_answer

__all__ = [
    'f1_score',
    'exact_match_score',
    'apply_no_ans_threshold',
    'make_eval_dict',
    'merge_eval',
    'find_all_best_thresh',
    'find_best_thresh',
    '_get_best_indexes',
    'get_final_text',
]


def _get_best_indexes(logits, n_best_size):
    """Get the n-best logits from a list."""
    index_and_score = sorted(enumerate(logits), key=lambda x: x[1], reverse=True)

    best_indexes = []
    for i in range(len(index_and_score)):
        if i >= n_best_size:
            break
        best_indexes.append(index_and_score[i][0])
    return best_indexes


def get_final_text(pred_text, orig_text, do_lower_case, verbose_logging=False):
    """Project the tokenized prediction back to the original text."""

    # When we created the data, we kept track of the alignment between original
    # (whitespace tokenized) tokens and our WordPiece tokenized tokens. So
    # now `orig_text` contains the span of our original text corresponding to
    # the span that we predicted.
    #
    # However, `orig_text` may contain extra characters that we don't want in
    # our prediction.
    #
    # For example, let's say:
    #   pred_text = steve smith
    #   orig_text = Steve Smith's
    #
    # We don't want to return `orig_text` because it contains the extra "'s".
    #
    # We don't want to return `pred_text` because it's already been normalized
    # (the SQuAD eval script also does punctuation stripping/lower casing but
    # our tokenizer does additional normalization like stripping accent
    # characters).
    #
    # What we really want to return is "Steve Smith".
    #
    # Therefore, we have to apply a semi-complicated alignment heuristic
    # between `pred_text` and `orig_text` to get a character-to-character
    # alignment. This can fail in certain cases in which case we just return
    # `orig_text`.

    def _strip_spaces(text):
        ns_chars = []
        ns_to_s_map = collections.OrderedDict()
        for (i, c) in enumerate(text):
            if c == " ":
                continue
            ns_to_s_map[len(ns_chars)] = i
            ns_chars.append(c)
        ns_text = "".join(ns_chars)
        return ns_text, ns_to_s_map

    # We first tokenize `orig_text`, strip whitespace from the result
    # and `pred_text`, and check if they are the same length. If they are
    # NOT the same length, the heuristic has failed. If they are the same
    # length, we assume the characters are one-to-one aligned.
    tokenizer = BasicTokenizer(do_lower_case=do_lower_case)

    tok_text = " ".join(tokenizer.tokenize(orig_text))

    start_position = tok_text.find(pred_text)
    if start_position == -1:
        if verbose_logging:
            logging.warning("Unable to find text: '%s' in '%s'" % (pred_text, orig_text))
        return orig_text
    end_position = start_position + len(pred_text) - 1

    (orig_ns_text, orig_ns_to_s_map) = _strip_spaces(orig_text)
    (tok_ns_text, tok_ns_to_s_map) = _strip_spaces(tok_text)

    if len(orig_ns_text) != len(tok_ns_text):
        if verbose_logging:
            logging.warning(
                "Length not equal after stripping spaces: '%s' vs '%s'", orig_ns_text, tok_ns_text,
            )
        return orig_text

    # We then project the characters in `pred_text` back to `orig_text` using
    # the character-to-character alignment.
    tok_s_to_ns_map = {}
    for (i, tok_index) in tok_ns_to_s_map.items():
        tok_s_to_ns_map[tok_index] = i

    orig_start_position = None
    if start_position in tok_s_to_ns_map:
        ns_start_position = tok_s_to_ns_map[start_position]
        if ns_start_position in orig_ns_to_s_map:
            orig_start_position = orig_ns_to_s_map[ns_start_position]

    if orig_start_position is None:
        if verbose_logging:
            logging.warning("Couldn't map start position")
        return orig_text

    orig_end_position = None
    if end_position in tok_s_to_ns_map:
        ns_end_position = tok_s_to_ns_map[end_position]
        if ns_end_position in orig_ns_to_s_map:
            orig_end_position = orig_ns_to_s_map[ns_end_position]

    if orig_end_position is None:
        if verbose_logging:
            logging.warning("Couldn't map end position")
        return orig_text

    output_text = orig_text[orig_start_position : (orig_end_position + 1)]
    return output_text


def f1_score(prediction, ground_truth):
    prediction_tokens = get_tokens(prediction)
    ground_truth_tokens = get_tokens(ground_truth)
    common = collections.Counter(prediction_tokens) & collections.Counter(ground_truth_tokens)
    num_same = sum(common.values())
    if len(ground_truth_tokens) == 0 or len(prediction_tokens) == 0:
        # If either is no-answer, then F1 is 1 if they agree, 0 otherwise
        return int(ground_truth_tokens == prediction_tokens)
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1


def exact_match_score(prediction, ground_truth):
    return int(normalize_answer(prediction) == normalize_answer(ground_truth))


def apply_no_ans_threshold(scores, na_probs, qid_to_has_ans, na_prob_thresh):
    new_scores = {}
    for qid, s in scores.items():
        pred_na = na_probs[qid] > na_prob_thresh
        if pred_na:
            new_scores[qid] = float(not qid_to_has_ans[qid])
        else:
            new_scores[qid] = s
    return new_scores


def make_eval_dict(exact_scores, f1_scores, qid_list=None):
    if not qid_list:
        total = len(exact_scores)
        return collections.OrderedDict(
            [
                ("exact", 100.0 * sum(exact_scores.values()) / total),
                ("f1", 100.0 * sum(f1_scores.values()) / total),
                ("total", total),
            ]
        )
    else:
        total = len(qid_list)
        return collections.OrderedDict(
            [
                ("exact", 100.0 * sum(exact_scores[k] for k in qid_list) / total),
                ("f1", 100.0 * sum(f1_scores[k] for k in qid_list) / total),
                ("total", total),
            ]
        )


def merge_eval(main_eval, new_eval, prefix):
    for k in new_eval:
        main_eval["%s_%s" % (prefix, k)] = new_eval[k]


def find_all_best_thresh(main_eval, preds, exact_raw, f1_raw, na_probs, qid_to_has_ans):
    best_exact, exact_thresh = find_best_thresh(preds, exact_raw, na_probs, qid_to_has_ans)
    best_f1, f1_thresh = find_best_thresh(preds, f1_raw, na_probs, qid_to_has_ans)

    main_eval["best_exact"] = best_exact
    main_eval["best_exact_thresh"] = exact_thresh
    main_eval["best_f1"] = best_f1
    main_eval["best_f1_thresh"] = f1_thresh


def find_best_thresh(preds, scores, na_probs, qid_to_has_ans):
    num_no_ans = sum(1 for k in qid_to_has_ans if not qid_to_has_ans[k])
    cur_score = num_no_ans
    best_score = cur_score
    best_thresh = 0.0
    qid_list = sorted(na_probs, key=lambda k: na_probs[k])
    for _, qid in enumerate(qid_list):
        if qid not in scores:
            continue
        if qid_to_has_ans[qid]:
            diff = scores[qid]
        else:
            if preds[qid]:
                diff = -1
            else:
                diff = 0
        cur_score += diff
        if cur_score > best_score:
            best_score = cur_score
            best_thresh = na_probs[qid]
    return 100.0 * best_score / len(scores), best_thresh



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.collections.nlp.data import BertPunctuationCapitalizationDataset
from nemo.collections.nlp.nm.data_layers.text_datalayer import TextDataLayer
from nemo.core import ChannelType, LabelsType, MaskType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['PunctuationCapitalizationDataLayer']


class PunctuationCapitalizationDataLayer(TextDataLayer):
    """
    Data layer for punctuation and capitalization.

    Args:
        text_file (str): file to sequences, each line should a sentence,
            No header.
        label_file (str): file to labels, each line corresponds to
            word labels for a sentence in the text_file. No header.
        tokenizer (TokenizerSpec): text tokenizer.
        max_seq_length (int): max sequence length minus 2 for [CLS] and [SEP]
        pad_label (str): ad value use for labels.
            by default, it's the neutral label.
        punct_label_ids (dict): 
            dict to map labels to label ids.
            Starts with pad_label->0 and then increases in alphabetical order
            For dev set use label_ids generated during training to support
            cases when not all labels are present in the dev set.
            For training set label_ids should be None.
        capit_label_ids (dict):
            dict to map labels to label ids.
            Starts with pad_label->0 and then increases in alphabetical order
            For dev set use label_ids generated during training to support
            cases when not all labels are present in the dev set.
            For training set label_ids should be None.
        num_samples (int):
            number of samples you want to use for the dataset.
                If -1, use all dataset. Useful for testing.
        shuffle (bool): whether to shuffle your data.
        batch_size (int): batch size
        ignore_extra_tokens (bool): whether to ignore extra tokens in
            the loss_mask
        ignore_start_end (bool):
            whether to ignore bos and eos tokens in the loss_mask
        use_cache (bool): whether to use data cache
        dataset_type (Dataset): Default BertPunctuationCapitalizationDataset.
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        input_ids:
            indices of tokens which constitute batches of masked text segments
        input_type_ids:
            tensor with 0's and 1's to denote the text segment type
        input_mask:
            bool tensor with 0s in place of tokens to be masked
        loss_mask:
            used to mask and ignore tokens in the loss function: indices of tokens which constitute batches of unmasked text segments
        subtokens_mask:
            used to mask all but the first subtoken of the work, could be useful during inference
        punct_labels: punctuation label ids
        capit_labels: capit_labels label ids
        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_type_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_mask": NeuralType(('B', 'T'), ChannelType()),
            "loss_mask": NeuralType(('B', 'T'), MaskType()),
            "subtokens_mask": NeuralType(('B', 'T'), ChannelType()),
            "punct_labels": NeuralType(('B', 'T'), LabelsType()),
            "capit_labels": NeuralType(('B', 'T'), LabelsType()),
        }

    def __init__(
        self,
        text_file,
        label_file,
        tokenizer,
        max_seq_length,
        pad_label='O',
        punct_label_ids=None,
        capit_label_ids=None,
        num_samples=-1,
        shuffle=False,
        batch_size=64,
        ignore_extra_tokens=False,
        ignore_start_end=False,
        overwrite_processed_files=False,
        num_workers=-1,
        pin_memory=False,
        dataset_type=BertPunctuationCapitalizationDataset,
    ):
        dataset_params = {
            'text_file': text_file,
            'label_file': label_file,
            'max_seq_length': max_seq_length,
            'tokenizer': tokenizer,
            'num_samples': num_samples,
            'pad_label': pad_label,
            'punct_label_ids': punct_label_ids,
            'capit_label_ids': capit_label_ids,
            'ignore_extra_tokens': ignore_extra_tokens,
            'ignore_start_end': ignore_start_end,
            'overwrite_processed_files': overwrite_processed_files,
        }
        super().__init__(
            dataset_type, dataset_params, batch_size, shuffle=shuffle, num_workers=num_workers, pin_memory=pin_memory
        )



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import torch
from torch.utils import data as pt_data

import nemo
from nemo.collections.nlp.data import TranslationDataset
from nemo.collections.nlp.nm.data_layers.text_datalayer import TextDataLayer
from nemo.core import ChannelType, LabelsType, MaskType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['TranslationDataLayer']


class TranslationDataLayer(TextDataLayer):
    """
    Data layer for neural machine translation from source (src) language to
    target (tgt) language.

    Args:
        tokenizer_src (TokenizerSpec): source language tokenizer
        tokenizer_tgt (TokenizerSpec): target language tokenizer
        dataset_src (str): path to source data
        dataset_tgt (str): path to target data
        tokens_in_batch (int): maximum allowed number of tokens in batches,
            batches will be constructed to minimize the use of <pad> tokens
        clean (bool): whether to use parallel data cleaning such as removing
            pairs with big difference in sentences length, removing pairs with
            the same tokens in src and tgt, etc; useful for training data layer
            and should not be used in evaluation data layer
        dataset_type (Dataset):
                the underlying dataset. Default: TranslationDataset
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        src_ids: indices of tokens which correspond to source sentences
        src_mask: bool tensor with 0s in place of source tokens to be masked
        tgt_ids: indices of tokens which correspond to target sentences
        tgt_mask: bool tensor with 0s in place of target tokens to be masked
        labels: indices of tokens which should be predicted from each of the
            corresponding target tokens in tgt_ids; for standard neural
            machine translation equals to tgt_ids shifted by 1 to the right
        sent_ids: indices of the sentences in a batch; important for
            evaluation with external metrics, such as SacreBLEU

        """
        return {
            "src_ids": NeuralType(('B', 'T'), ChannelType()),
            "src_mask": NeuralType(('B', 'T'), ChannelType()),
            "tgt_ids": NeuralType(('B', 'T'), LabelsType()),
            "tgt_mask": NeuralType(('B', 'T'), MaskType()),
            "labels": NeuralType(('B', 'T'), LabelsType()),
            "sent_ids": NeuralType(tuple('B'), ChannelType()),
        }

    def __init__(
        self,
        tokenizer_src,
        tokenizer_tgt,
        dataset_src,
        dataset_tgt,
        tokens_in_batch=1024,
        shuffle=False,
        clean=False,
        dataset_type=TranslationDataset,
    ):
        dataset_params = {
            'tokenizer_src': tokenizer_src,
            'tokenizer_tgt': tokenizer_tgt,
            'dataset_src': dataset_src,
            'dataset_tgt': dataset_tgt,
            'tokens_in_batch': tokens_in_batch,
            'clean': clean,
        }
        super().__init__(dataset_type, dataset_params, batch_size=1, shuffle=shuffle)

        if self._placement == nemo.core.DeviceType.AllGpu:
            sampler = pt_data.distributed.DistributedSampler(self._dataset)
        else:
            sampler = None

        self._dataloader = pt_data.DataLoader(
            dataset=self._dataset, batch_size=1, collate_fn=self._collate_fn, shuffle=sampler is None, sampler=sampler
        )

    def _collate_fn(self, x):
        src_ids, src_mask, tgt_ids, tgt_mask, labels, sent_ids = x[0]
        src_ids = torch.Tensor(src_ids).long().to(self._device)
        src_mask = torch.Tensor(src_mask).float().to(self._device)
        tgt_ids = torch.Tensor(tgt_ids).long().to(self._device)
        tgt_mask = torch.Tensor(tgt_mask).float().to(self._device)
        labels = torch.Tensor(labels).long().to(self._device)
        sent_ids = torch.Tensor(sent_ids).long().to(self._device)
        return src_ids, src_mask, tgt_ids, tgt_mask, labels, sent_ids

    @property
    def dataset(self):
        return None

    @property
    def data_iterator(self):
        return self._dataloader



import torch
from torch import nn

from nemo.backends.pytorch.nm import LossNM
from nemo.core.neural_types import LabelsType, LogitsType, LossType, MaskType, NeuralType, RegressionValuesType
from nemo.utils.decorators import add_port_docs

__all__ = ['SequenceLoss', 'CrossEntropyLossNM', 'MSELoss', 'LossAggregatorNM', 'BCEWithLogitsLossNM']


class SequenceLoss(LossNM):
    """Loss for seq2seq tasks

    Args:
        pad_id (int): Label position of padding symbol.
            Defaults to 0.
        smoothing_coef (float): Label smoothing coefficient in range [0, 1].
            Defaults to 0.0.
        sample_wise (bool): Flag indicates if loss sum divisor should be batch
            size.
            Defaults to False.
        aux_ctc (bool): Whether to add auxiliary CTC loss.
            Defaults to False.
        ctc_initial_coef (float): Initial coefficient to multiply ctc component
            by.
            Defaults to 0.1.
        ctc_blank_id (int): ID of blank symbols to pass to mask when
            calculating ctc loss.
            Defaults to None.
        eps (float): small number to prevent division by zero in loss calculation
            Defaults to 1e-5.

    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {'log_probs': NeuralType(axes=('B', 'T', 'D')), 'targets': NeuralType(axes=('B', 'T'))}

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(
        self,
        pad_id=0,
        smoothing_coef=0.0,
        sample_wise=False,
        aux_ctc=False,
        ctc_initial_coef=0.1,
        ctc_blank_id=None,
        eps=1e-5,
    ):
        assert (not aux_ctc) or (ctc_blank_id is not None), "Should be a blank id if using CTC loss"

        super().__init__()

        self.pad_id = pad_id
        self.smoothing_coef = smoothing_coef
        self.sample_wise = sample_wise
        self.aux_ctc = aux_ctc
        self.ctc_coef = ctc_initial_coef
        self.eps = eps

        if aux_ctc:
            self.ctc = nn.CTCLoss(blank=ctc_blank_id, reduction='none', zero_infinity=True)
            self.ctc = self.ctc.to(self._device)

    def _loss_function(self, log_probs, targets):
        """(BTC, BT) -> 0"""

        pad_mask = (targets != self.pad_id).long()
        loss = self._ce_loss(log_probs, targets, pad_mask)

        if self.aux_ctc:
            ctc_loss = self._ctc_loss(log_probs, targets, pad_mask)
            loss += self.ctc_coef * ctc_loss

        assert loss.dim() == 0, "Zero-dim tensor check"

        return loss

    def _ce_loss(self, log_probs, targets, pad_mask):
        target_log_probs = log_probs.gather(2, targets.unsqueeze(2)).squeeze(2)
        loss = (1.0 - self.smoothing_coef) * target_log_probs + self.smoothing_coef * log_probs.mean(-1)
        pad_mask = pad_mask.float()
        loss = -torch.sum(loss * pad_mask)
        if self.sample_wise:
            loss /= target_log_probs.size(0)
        else:
            loss /= pad_mask.sum() + self.eps
        return loss

    def _ctc_loss(self, log_probs, targets, pad_mask):
        lengths = pad_mask.sum(-1)
        loss = self.ctc(log_probs.transpose(0, 1), targets, lengths, lengths)
        loss = torch.mean(loss)
        return loss


class CrossEntropyLossNM(LossNM):
    """
    CrossEntropyLoss
    Args:
        logits_ndim (int): number of dimensions (or rank) of the logits tensor
        weight (list): list of rescaling weight given to each class
        reduction (str): type of the reduction over the batch
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            "logits": NeuralType(['B'] + ['ANY'] * (self._logits_dim - 1), LogitsType()),
            "labels": NeuralType(['B'] + ['ANY'] * (self._logits_dim - 2), LabelsType()),
            "loss_mask": NeuralType(['B'] + ['ANY'] * (self._logits_dim - 2), MaskType(), optional=True),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        loss:
            NeuralType(None)
        """
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(self, logits_ndim=2, weight=None, reduction='mean'):
        super().__init__()

        if weight:
            weight = torch.FloatTensor(weight).to(self._device)
        self._criterion = nn.CrossEntropyLoss(weight=weight, reduction=reduction)
        self._logits_dim = logits_ndim

    def _loss_function(self, logits, labels, loss_mask=None):
        """
        Args:
            logits (float): output of the classifier
            labels (long): ground truth labels
            loss_mask (bool/float/int): tensor to specify the masking
        """
        logits_flatten = torch.flatten(logits, start_dim=0, end_dim=-2)
        labels_flatten = torch.flatten(labels, start_dim=0, end_dim=-1)

        if loss_mask is not None:
            if loss_mask.dtype is not torch.bool:
                loss_mask = loss_mask > 0.5
            loss_mask_flatten = torch.flatten(loss_mask, start_dim=0, end_dim=-1)
            logits_flatten = logits_flatten[loss_mask_flatten]
            labels_flatten = labels_flatten[loss_mask_flatten]

        if len(labels_flatten) == 0:
            return self._criterion(logits, torch.argmax(logits, dim=-1))

        loss = self._criterion(logits_flatten, labels_flatten)
        return loss


class MSELoss(LossNM):
    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.

        preds:
            0: AxisType(RegressionTag)

        labels:
            0: AxisType(RegressionTag)
        """
        return {
            "preds": NeuralType(tuple('B'), RegressionValuesType()),
            "labels": NeuralType(tuple('B'), LabelsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        loss:
            NeuralType(None)
        """
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(self, reduction='mean'):
        super().__init__()
        self._criterion = nn.MSELoss(reduction=reduction)

    def _loss_function(self, preds, labels):
        loss = self._criterion(preds, labels)
        return loss


class LossAggregatorNM(LossNM):
    """
    Neural module which combines sums several losses into one.

    Args:
        num_inputs (int): number of input losses
        weights (list of floats): a list of coefficient for merging losses
    """

    @property
    def input_ports(self):
        """Returns definitions of module input ports.

        """
        input_ports = {}
        for i in range(self._num_losses):
            input_ports["loss_" + str(i + 1)] = NeuralType(elements_type=LossType())

        return input_ports

    @property
    def output_ports(self):
        """Returns definitions of module output ports.

        loss:
            NeuralType(None)
        """
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(self, num_inputs=2, weights=None):
        # Store number of inputs/losses.
        self._num_losses = num_inputs
        if weights is not None and len(weights) != num_inputs:
            raise ValueError("Length of weights should be equal to the number of inputs (num_inputs)")

        self._weights = weights
        LossNM.__init__(self)

    def _loss_function(self, **kwargs):
        values = [kwargs[x] for x in sorted(kwargs.keys())]
        loss = torch.zeros_like(values[0])
        for loss_idx, loss_value in enumerate(values):
            if self._weights is not None:
                loss = loss.add(loss_value, alpha=self._weights[loss_idx])
            else:
                loss = loss.add(loss_value)
        return loss


class BCEWithLogitsLossNM(LossNM):
    """
    CrossEntropyLoss
    Args:
        logits_ndim (int): number of dimensions (or rank) of the logits tensor
        weight (list): list of rescaling weight given to each class
        reduction (str): type of the reduction over the batch
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            "logits": NeuralType(['B'] + ['ANY'] * (self._logits_dim - 1), LogitsType()),
            "labels": NeuralType(['B'] + ['ANY'] * (self._logits_dim - 2), LabelsType()),
            "loss_mask": NeuralType(['B'] + ['ANY'] * (self._logits_dim - 2), MaskType(), optional=True),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        loss:
            NeuralType(None)
        """
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(self, logits_ndim=2, weight=None, reduction='mean'):
        super().__init__()

        if weight:
            weight = torch.FloatTensor(weight).to(self._device)
        self._criterion = nn.BCEWithLogitsLoss(weight=weight, reduction=reduction)
        self._logits_dim = logits_ndim

    def _loss_function(self, logits, labels, loss_mask=None):
        """
        Args:
            logits (float): output of the classifier
            labels (long): ground truth labels
            loss_mask (bool/float/int): tensor to specify the masking
        """
        logits_flatten = torch.flatten(logits, start_dim=0, end_dim=-2)
        labels_flatten = torch.flatten(labels, start_dim=0, end_dim=-1)

        if loss_mask is not None:
            if loss_mask.dtype is not torch.bool:
                loss_mask = loss_mask > 0.5
            loss_mask_flatten = torch.flatten(loss_mask, start_dim=0, end_dim=-1)
            logits_flatten = logits_flatten[loss_mask_flatten]
            labels_flatten = labels_flatten[loss_mask_flatten]

        if len(labels_flatten) == 0:
            return 0

        loss = self._criterion(logits_flatten, labels_flatten)
        return loss



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2018 The Google AI Language Team Authors and
# The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Utility functions for Token Classification NLP tasks
Some parts of this code were adapted from the HuggingFace library at
https://github.com/huggingface/pytorch-pretrained-BERT
"""

import itertools
import os
import pickle

import numpy as np
from torch.utils.data import Dataset

from nemo import logging
from nemo.collections.nlp.data.datasets.datasets_utils.data_preprocessing import get_label_stats, get_stats

__all__ = ['BertTokenClassificationDataset', 'BertTokenClassificationInferDataset']


def get_features(
    queries,
    max_seq_length,
    tokenizer,
    label_ids=None,
    pad_label='O',
    raw_labels=None,
    ignore_extra_tokens=False,
    ignore_start_end=False,
):
    """
    Args:
    queries (list of str): text sequences
    max_seq_length (int): max sequence length minus 2 for [CLS] and [SEP]
    tokenizer (Tokenizer): such as NemoBertTokenizer
    pad_label (str): pad value use for labels.
        by default, it's the neutral label.
    raw_labels (list of str): list of labels for every word in a sequence
    label_ids (dict): dict to map labels to label ids. Starts
        with pad_label->0 and then increases in alphabetical order.
        Required for training and evaluation, not needed for inference.
    ignore_extra_tokens (bool): whether to ignore extra tokens in
        the loss_mask,
    ignore_start_end (bool): whether to ignore bos and eos tokens in
        the loss_mask
    """
    all_subtokens = []
    all_loss_mask = []
    all_subtokens_mask = []
    all_segment_ids = []
    all_input_ids = []
    all_input_mask = []
    sent_lengths = []
    all_labels = []
    with_label = False

    if raw_labels is not None:
        with_label = True

    for i, query in enumerate(queries):
        words = query.strip().split()

        # add bos token
        subtokens = [tokenizer.cls_token]
        loss_mask = [1 - ignore_start_end]
        subtokens_mask = [0]
        if with_label:
            pad_id = label_ids[pad_label]
            labels = [pad_id]
            query_labels = [label_ids[lab] for lab in raw_labels[i]]

        for j, word in enumerate(words):
            word_tokens = tokenizer.text_to_tokens(word)
            subtokens.extend(word_tokens)

            loss_mask.append(1)
            loss_mask.extend([int(not ignore_extra_tokens)] * (len(word_tokens) - 1))

            subtokens_mask.append(1)
            subtokens_mask.extend([0] * (len(word_tokens) - 1))

            if with_label:
                labels.extend([query_labels[j]] * len(word_tokens))
        # add eos token
        subtokens.append(tokenizer.sep_token)
        loss_mask.append(1 - ignore_start_end)
        subtokens_mask.append(0)
        sent_lengths.append(len(subtokens))
        all_subtokens.append(subtokens)
        all_loss_mask.append(loss_mask)
        all_subtokens_mask.append(subtokens_mask)
        all_input_mask.append([1] * len(subtokens))

        if with_label:
            labels.append(pad_id)
            all_labels.append(labels)

    max_seq_length = min(max_seq_length, max(sent_lengths))
    logging.info(f'Max length: {max_seq_length}')
    get_stats(sent_lengths)
    too_long_count = 0

    for i, subtokens in enumerate(all_subtokens):
        if len(subtokens) > max_seq_length:
            subtokens = [tokenizer.cls_token] + subtokens[-max_seq_length + 1 :]
            all_input_mask[i] = [1] + all_input_mask[i][-max_seq_length + 1 :]
            all_loss_mask[i] = [int(not ignore_start_end)] + all_loss_mask[i][-max_seq_length + 1 :]
            all_subtokens_mask[i] = [0] + all_subtokens_mask[i][-max_seq_length + 1 :]

            if with_label:
                all_labels[i] = [pad_id] + all_labels[i][-max_seq_length + 1 :]
            too_long_count += 1

        all_input_ids.append(tokenizer.tokens_to_ids(subtokens))

        if len(subtokens) < max_seq_length:
            extra = max_seq_length - len(subtokens)
            all_input_ids[i] = all_input_ids[i] + [0] * extra
            all_loss_mask[i] = all_loss_mask[i] + [0] * extra
            all_subtokens_mask[i] = all_subtokens_mask[i] + [0] * extra
            all_input_mask[i] = all_input_mask[i] + [0] * extra

            if with_label:
                all_labels[i] = all_labels[i] + [pad_id] * extra

        all_segment_ids.append([0] * max_seq_length)

    logging.warning(f'{too_long_count} are longer than {max_seq_length}')

    for i in range(min(len(all_input_ids), 5)):
        logging.info("*** Example ***")
        logging.info("i: %s", i)
        logging.info("subtokens: %s", " ".join(list(map(str, all_subtokens[i]))))
        logging.info("loss_mask: %s", " ".join(list(map(str, all_loss_mask[i]))))
        logging.info("input_mask: %s", " ".join(list(map(str, all_input_mask[i]))))
        logging.info("subtokens_mask: %s", " ".join(list(map(str, all_subtokens_mask[i]))))
        if with_label:
            logging.info("labels: %s", " ".join(list(map(str, all_labels[i]))))
    return (all_input_ids, all_segment_ids, all_input_mask, all_loss_mask, all_subtokens_mask, all_labels)


class BertTokenClassificationDataset(Dataset):
    """
    Creates dataset to use during training for token classification
    tasks with a pretrained model.

    Converts from raw data to an instance that can be used by
    NMDataLayer.

    For dataset to use during inference without labels, see
    BertTokenClassificationInferDataset.

    Args:
        text_file (str): file to sequences, each line should a sentence,
            No header.
        label_file (str): file to labels, each line corresponds to
            word labels for a sentence in the text_file. No header.
        max_seq_length (int): max sequence length minus 2 for [CLS] and [SEP]
        tokenizer (Tokenizer): such as NemoBertTokenizer
        num_samples (int): number of samples you want to use for the dataset.
            If -1, use all dataset. Useful for testing.
        pad_label (str): pad value use for labels.
            by default, it's the neutral label.
        label_ids (dict): label_ids (dict): dict to map labels to label ids.
            Starts with pad_label->0 and then increases in alphabetical order
            For dev set use label_ids generated during training to support
            cases when not all labels are present in the dev set.
            For training set label_ids should be None.
        ignore_extra_tokens (bool): whether to ignore extra tokens in
            the loss_mask,
        ignore_start_end (bool): whether to ignore bos and eos tokens in
            the loss_mask
    """

    def __init__(
        self,
        text_file,
        label_file,
        max_seq_length,
        tokenizer,
        num_samples=-1,
        pad_label='O',
        label_ids=None,
        ignore_extra_tokens=False,
        ignore_start_end=False,
        use_cache=False,
    ):

        if use_cache:
            # Cache features
            data_dir = os.path.dirname(text_file)
            filename = os.path.basename(text_file)

            if not filename.endswith('.txt'):
                raise ValueError("{text_file} should have extension .txt")

            tokenizer_type = type(tokenizer.tokenizer).__name__
            vocab_size = getattr(tokenizer, "vocab_size", 0)
            features_pkl = os.path.join(
                data_dir, "cached_{}_{}_{}_{}".format(filename, tokenizer_type, str(max_seq_length), str(vocab_size)),
            )
            label_ids_pkl = os.path.join(data_dir, "label_ids.pkl")

        if use_cache and os.path.exists(features_pkl) and os.path.exists(label_ids_pkl):
            # If text_file was already processed, load from pickle
            features = pickle.load(open(features_pkl, 'rb'))
            logging.info(f'features restored from {features_pkl}')

            label_ids = pickle.load(open(label_ids_pkl, 'rb'))
            logging.info(f'Labels to ids dict restored from {label_ids_pkl}')
        else:
            if num_samples == 0:
                raise ValueError("num_samples has to be positive", num_samples)

            with open(text_file, 'r') as f:
                text_lines = f.readlines()

            # Collect all possible labels
            unique_labels = set()
            labels_lines = []
            with open(label_file, 'r') as f:
                for line in f:
                    line = line.strip().split()
                    labels_lines.append(line)
                    unique_labels.update(line)

            if len(labels_lines) != len(text_lines):
                raise ValueError("Labels file should contain labels for every word")

            if num_samples > 0:
                dataset = list(zip(text_lines, labels_lines))
                dataset = dataset[:num_samples]

                dataset = list(zip(*dataset))
                text_lines = dataset[0]
                labels_lines = dataset[1]

            # for dev/test sets use label mapping from training set
            if label_ids:
                if len(label_ids) != len(unique_labels):
                    logging.warning(
                        f'Not all labels from the specified'
                        + ' label_ids dictionary are present in the'
                        + ' current dataset. Using the provided'
                        + ' label_ids dictionary.'
                    )
                else:
                    logging.info(f'Using the provided label_ids dictionary.')
            else:
                logging.info(
                    f'Creating a new label to label_id dictionary.'
                    + ' It\'s recommended to use label_ids generated'
                    + ' during training for dev/test sets to avoid'
                    + ' errors if some labels are not'
                    + ' present in the dev/test sets.'
                    + ' For training set label_ids should be None.'
                )

                label_ids = {pad_label: 0}
                if pad_label in unique_labels:
                    unique_labels.remove(pad_label)
                for label in sorted(unique_labels):
                    label_ids[label] = len(label_ids)

            features = get_features(
                text_lines,
                max_seq_length,
                tokenizer,
                pad_label=pad_label,
                raw_labels=labels_lines,
                label_ids=label_ids,
                ignore_extra_tokens=ignore_extra_tokens,
                ignore_start_end=ignore_start_end,
            )

            if use_cache:
                pickle.dump(features, open(features_pkl, "wb"))
                logging.info(f'features saved to {features_pkl}')

                pickle.dump(label_ids, open(label_ids_pkl, "wb"))
                logging.info(f'labels to ids dict saved to {label_ids_pkl}')

        self.all_input_ids = features[0]
        self.all_segment_ids = features[1]
        self.all_input_mask = features[2]
        self.all_loss_mask = features[3]
        self.all_subtokens_mask = features[4]
        self.all_labels = features[5]
        self.label_ids = label_ids

        infold = text_file[: text_file.rfind('/')]
        merged_labels = itertools.chain.from_iterable(self.all_labels)
        logging.info('Three most popular labels')
        _, self.label_frequencies, _ = get_label_stats(merged_labels, infold + '/label_stats.tsv')

        # save label_ids
        out = open(infold + '/label_ids.csv', 'w')
        labels, _ = zip(*sorted(self.label_ids.items(), key=lambda x: x[1]))
        out.write('\n'.join(labels))
        logging.info(f'Labels: {self.label_ids}')
        logging.info(f'Labels mapping saved to : {out.name}')

    def __len__(self):
        return len(self.all_input_ids)

    def __getitem__(self, idx):
        return (
            np.array(self.all_input_ids[idx]),
            np.array(self.all_segment_ids[idx]),
            np.array(self.all_input_mask[idx], dtype=np.long),
            np.array(self.all_loss_mask[idx]),
            np.array(self.all_subtokens_mask[idx]),
            np.array(self.all_labels[idx]),
        )


class BertTokenClassificationInferDataset(Dataset):
    """
    Creates dataset to use during inference for token classification
    tasks with a pretrained model.

    Converts from raw data to an instance that can be used by
    NMDataLayer.

    For dataset to use during training with labels, see
    BertTokenClassificationDataset.

    Args:
        queries (list): list of queries to run inference on
        max_seq_length (int): max sequence length minus 2 for [CLS] and [SEP]
        tokenizer (Tokenizer): such as NemoBertTokenizer
    """

    def __init__(self, queries, max_seq_length, tokenizer):
        features = get_features(queries, max_seq_length, tokenizer)

        self.all_input_ids = features[0]
        self.all_segment_ids = features[1]
        self.all_input_mask = features[2]
        self.all_loss_mask = features[3]
        self.all_subtokens_mask = features[4]

    def __len__(self):
        return len(self.all_input_ids)

    def __getitem__(self, idx):
        return (
            np.array(self.all_input_ids[idx]),
            np.array(self.all_segment_ids[idx]),
            np.array(self.all_input_mask[idx], dtype=np.long),
            np.array(self.all_loss_mask[idx]),
            np.array(self.all_subtokens_mask[idx]),
        )



# Copyright (c) 2019 NVIDIA Corporation
"""Core PyTorch-base Neural Modules"""
__all__ = [
    'SequenceEmbedding',
    'ZerosLikeNM',
]

from typing import Dict, Iterable, Mapping, Optional, Set

import torch
import torch.nn as nn

from nemo.backends.pytorch.nm import NonTrainableNM, TrainableNM
from nemo.core.neural_types import *


class SequenceEmbedding(TrainableNM):
    @property
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        # return {"input_seq": NeuralType({0: AxisType(TimeTag), 1: AxisType(BatchTag)})}
        return {"input_seq": NeuralType(('B', 'T'))}

    @property
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        # return {"outputs": NeuralType({0: AxisType(TimeTag), 1: AxisType(BatchTag), 2: AxisType(ChannelTag),})}
        return {"outputs": NeuralType(('B', 'T', 'C'), ChannelType())}

    def __init__(self, voc_size, hidden_size, dropout=0.0):
        super().__init__()

        self.voc_size = voc_size
        self.hidden_size = hidden_size
        self.dropout = dropout
        self.embedding = nn.Embedding(self.voc_size, self.hidden_size)
        if self.dropout != 0.0:
            self.embedding_dropout = nn.Dropout(self.dropout)
        self.to(self._device)

    def forward(self, input_seq):
        embedded = self.embedding(input_seq)
        if self.dropout != 0.0:
            embedded = self.embedding_dropout(embedded)
        return embedded


class ZerosLikeNM(NonTrainableNM):
    @property
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        # return {"input_type_ids": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag),})}
        return {"input_type_ids": NeuralType(('B', 'T'), VoidType())}

    @property
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        # return {"input_type_ids": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag),})}
        return {"input_type_ids": NeuralType(('B', 'T'), ChannelType())}

    def __init__(self):
        super().__init__()

    def forward(self, input_type_ids):
        return torch.zeros_like(input_type_ids).long()



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.collections.nlp.nm.trainables.common.huggingface.albert_nm import Albert
from nemo.collections.nlp.nm.trainables.common.huggingface.bert_nm import BERT
from nemo.collections.nlp.nm.trainables.common.huggingface.roberta_nm import Roberta

__all__ = ['MODELS', 'get_huggingface_lm_model', 'get_huggingface_lm_models_list']


def get_huggingface_lm_model(pretrained_model_name, bert_config=None):
    '''
    Returns the dict of special tokens associated with the model.
    Args:
    pretrained_mode_name ('str'): name of the pretrained model from the hugging face list,
        for example: bert-base-cased
    bert_config: path to model configuration file.
    '''
    model_type = pretrained_model_name.split('-')[0]
    if model_type in MODELS:
        if bert_config:
            return MODELS[model_type]['class'](config_filename=bert_config)
        else:
            return MODELS[model_type]['class'](pretrained_model_name=pretrained_model_name)
    else:
        raise ValueError(f'{pretrained_model_name} is not supported')


MODELS = {
    'bert': {'default': 'bert-base-uncased', 'class': BERT},
    'roberta': {'default': 'roberta-base', 'class': Roberta},
    'albert': {'default': 'albert-base-v2', 'class': Albert},
}


def get_huggingface_lm_models_list():
    '''
    Returns the list of supported HuggingFace models
    '''
    huggingface_models = []
    for model in MODELS:
        model_names = [x.pretrained_model_name for x in MODELS[model]['class'].list_pretrained_models()]
        huggingface_models.extend(model_names)
    return huggingface_models
